import pathlib
import json
import numpy as np

from .serialisable import NumpyEncoder

def check_create(file_path: pathlib.Path):

    if not file_path.exists():
        paths_to_create = [file_path]
        while not paths_to_create[0].parent.exists():
            paths_to_create.insert(0, paths_to_create[0].parent)
        [f.mkdir() for f in paths_to_create if not f.suffix]
        return False
    return True

def save_npy(path, array):
    with open(path, 'wb+') as f:
        np.save(f, array, allow_pickle=False)


def open_npy(path):
    with open(path, 'rb') as f:
        return np.load(f)


def save_json(filename, save_dict):
    with open(filename, 'w') as f:
        json.dump(save_dict, f, cls=NumpyEncoder, indent=4,
                  separators=(',', ': '))
        
def open_json(filename):
    with open(filename, 'r') as f:
        return json.load(f)



import pandas as pd
from sqlalchemy import create_engine, func, and_, case, MetaData, select, column
from sqlalchemy.orm import sessionmaker, aliased
from itertools import groupby
from optuna.storages._rdb.models import StudyModel, TrialModel, TrialValueModel, TrialUserAttributeModel
from matplotlib.font_manager import FontProperties
from typing import List, Dict, Tuple, Union

from ..logger import get_default_storage, Experiment, Tag, Component, DataPoint, ExpMetadata, Evaluation, DataPointInfo, get_coverages

from matplotlib import pyplot as plt
import numpy as np
import seaborn as sns
from scipy import interpolate
import matplotlib.colors as mcolors
import os
import optuna
import shap
import pathlib

from collections import Counter

from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error, r2_score

PALETTE_STRING = 'rocket'

def get_data_for_experiments_tag_and_component(experiments, tag_name, component_name, db_url=None):
    # Create a DB session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    
    try:
        # Ensure experiment_numbers is a list
        if not isinstance(experiments, list):
            experiments = [experiments]
        allvalues = []
        timestamps = []
        # Query and plot for each experiment number
        for experiment in experiments:
            if isinstance(experiment, str):
                tempexp = experiment
                experiment = session.query(Experiment).filter_by(name=experiment).first()
            tag = session.query(Tag).filter_by(name=tag_name).first()
            component = session.query(Component).filter_by(name=component_name).first()

            
            if not experiment or not tag or not component:
                print(f"Experiment: {tempexp}, Tag {tag}, or Component {component} not found!")
                continue
            
            data_points = (
                session.query(DataPoint)
                .filter_by(experiment_id=experiment.experiment_id, tag_id=tag.tag_id, component_id=component.component_id)
                .order_by(DataPoint.step)
                .all()
            )
            
            # Extract the data for plotting
            timestamps.append([dp.step for dp in data_points])

            def determine_tag_type(dp):
                float_value = dp.float_value
                if float_value is not None:
                    return 'float_value'
                int_value = dp.int_value
                if int_value is not None:
                    return 'int_value'
                bool_value = dp.bool_value
                if bool_value is not None:
                    return 'bool_value'
                string_value = dp.string_value
                if string_value is not None:
                    return 'string_value'
                
            values = [getattr(dp, determine_tag_type(dp)) for dp in data_points]  # assuming you want to plot float values
            allvalues.append(values)

    except Exception as e:
        print(f"An error occurred: {str(e)}")
        
    finally:
        session.close()
        engine.dispose()

    return experiments, timestamps, allvalues

def plot_data_for_experiments_tag_and_component(experiments, tag_name, component_name, db_url=None):
   
    plt.figure(figsize=(10, 6))
    
    experiments, timestamps, allvalues = get_data_for_experiments_tag_and_component(experiments, tag_name, component_name, db_url=db_url)
    for experiment, timestamps, values in zip(experiments, timestamps, allvalues):
        plt.plot(timestamps, values, marker='o', linestyle='-', label=f"Experiment {experiment.split('_')[-1]}")
    
    # Finalize the plot
    plt.title(f'Data for Tag: {tag_name}, Component: {component_name}')
    plt.xlabel('Timestamp')
    plt.ylabel('Value')
    plt.legend()
    plt.grid(True)
    plt.show()
        


def find_exps(env, reward_type, basename, db_url=None):
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Create aliases for the metadata table
    env_metadata = aliased(ExpMetadata)
    reward_metadata = aliased(ExpMetadata)

    # Query
    experiments = session.query(Experiment).join(
        env_metadata, Experiment.experiment_id == env_metadata.experiment_id
    ).join(
        reward_metadata, Experiment.experiment_id == reward_metadata.experiment_id
    ).filter(
        and_(env_metadata.key == 'env', env_metadata.value == env),
        and_(reward_metadata.key == 'reward_type', reward_metadata.value == reward_type),
        Experiment.name.like(f"{basename}_%")
    ).all()

    session.close()
    return experiments

def plot_evaluations_for_experiment(experiment_names, db_url=None):
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Initialize lists to store step and success rate values
    steps = []
    success_rates = []
    if not isinstance(experiment_names, list):
        experiment_names = [experiment_names]
    # Loop through each experiment name
    for experiment_name in experiment_names:
        # Query for the experiment ID using the experiment name
        experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
        
        if experiment:
            success_rate, step = get_coverages(experiment.name)
            steps.append(step)
            success_rates.append(success_rate)
    # Close the database session
    session.close()
    engine.dispose()
    return experiment_names, steps, success_rates
    # Create a plot for each experiment
    for experiment_name, step_values, success_rate_values in zip(experiment_names, steps, success_rates):
        plt.plot(step_values, success_rate_values, label=experiment_name)

    # Add labels and legend
    plt.xlabel('Step')
    plt.ylabel('Success Rate')
    plt.legend()

    # Show the plot
    plt.show()

    


def get_best_experiments_in_study(study_name:str, cut_off_value: float, db_url=None):
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    columns_to_select = [
        StudyModel.study_name,
        TrialModel.number,
        TrialValueModel.value,
        TrialUserAttributeModel.value_json,
    ]
    # Define your custom SQL query using explicit column expressions with column function
    query = (
        session.query(*columns_to_select)
        .join(TrialModel, StudyModel.study_id == TrialModel.study_id)
        .join(TrialValueModel, TrialModel.trial_id == TrialValueModel.trial_id)
        .join(TrialUserAttributeModel, TrialModel.trial_id == TrialUserAttributeModel.trial_id)
        .filter(and_(StudyModel.study_name == study_name, TrialValueModel.value >= cut_off_value))
        .order_by(TrialValueModel.value.desc())
    )

    results = query.all()
    session.close()
    engine.dispose()
    results.sort(key=lambda x: x[2])  # Sort by TrialValueModel.value
    grouped_results = []
    for value, group in groupby(results, key=lambda x: x[2]):
        group_list = []
        for result in group:
            study_name, trial_number, value, user_attribute_value = result
            group_list.append([study_name, trial_number, value, user_attribute_value.strip('"')])
        grouped_results.append(group_list)
    return grouped_results

def get_all_experiments_in_study(study_name:str, db_url=None):
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    columns_to_select = [
        StudyModel.study_name,
        TrialUserAttributeModel.value_json,
    ]
    # Define your custom SQL query using explicit column expressions with column function
    query = (
        session.query(*columns_to_select)
        .join(TrialModel, StudyModel.study_id == TrialModel.study_id)
        .join(TrialUserAttributeModel, TrialModel.trial_id == TrialUserAttributeModel.trial_id)
        .filter(StudyModel.study_name == study_name)
    )

    results = query.all()
    session.close()
    engine.dispose()
    return [x[1].strip('"') for x in results]

def moving_average(data, window_size):
    return np.convolve(data, np.ones(window_size)/window_size, mode='valid')

def plot_with_smoothing(experiments, timestamps, allvalues, smoothing_window=10):
    
    avg_timestamps = [moving_average(t, smoothing_window) for t in timestamps]
    avg_values = [moving_average(v, smoothing_window) for v in allvalues]
    plot_data(experiments, 'Steps', avg_timestamps, 'Episodic Reward', avg_values, xlims=[min(timestamps[0]), max(timestamps[0])])

def plot_data(experiments, xlabel, xvalues, ylabel, yvalues, ylims=[], xlims=[], legend = True, file_path: Union[str, pathlib.Path] = None):
    set_sns_plot_settings(18,18,18)

    pallete = sns.color_palette(PALETTE_STRING, len(experiments))
    fig = plt.figure(figsize=(8, 6))
    for idx, (experiment, x_value, y_value) in enumerate(zip(experiments, xvalues, yvalues)):
            sns.lineplot(x=x_value, y=y_value, linestyle='-', linewidth=3, alpha=0.7, label=f"{experiment}", color = pallete[idx])
    # Finalize the plot
    if ylims:
        plt.ylim(ylims)
    if xlims:
        plt.xlim(xlims)
    plt.xlabel(xlabel, fontweight='bold')
    plt.ylabel(ylabel, fontweight='bold')
    
    if legend:
        plt.legend()
    else:
        plt.legend().remove()
    
    plt.grid(True)
    plt.xticks(rotation=-10)
    plt.tight_layout()
    if file_path:
        file_path = convert_fp_to_pathlib_create_directories(file_path)
        plt.savefig(file_path)
        plt.close() 
    plt.show()

def scatter_plot_data(experiments, xlabel, xvalues, ylabel, yvalues, ylims=[], xlims=[], leg_loc='best', show=True):
    set_sns_plot_settings()
    pallete = sns.color_palette(PALETTE_STRING, len(experiments))
    # fig = plt.figure(figsize=(8, 6))
    plt.figure(figsize=(8, 6)) 
    for idx, (experiment, x_value, y_value) in enumerate(zip(experiments, xvalues, yvalues)):
            sns.scatterplot(x=x_value, y=y_value, alpha=0.4, label=f"{experiment}", color = pallete[idx])
    # Finalize the plot
    if ylims:
        plt.ylim(ylims)
    if xlims:
        plt.xlim(xlims)
    plt.xlabel(xlabel, fontweight='bold')
    plt.ylabel(ylabel, fontweight='bold')
    plt.legend(loc=leg_loc)
    plt.grid(True)
    plt.xticks(rotation=-10)
    if show: 
        plt.show()

def plot_interpolated_studies_episodic_reward(study_names, agent_name, environment_name, mov_avg=None):
    # Plotting
    set_sns_plot_settings()
    plt.figure(figsize=(10, 6))
    pallete = sns.color_palette(PALETTE_STRING, len(study_names))
    for i,(study, agent) in enumerate(zip(study_names, agent_name)):
        exps = get_all_experiments_in_study(study)
        _ , timestamps, allvalues = get_data_for_experiments_tag_and_component(exps,'train/ep_rew', agent)
        if mov_avg:
            allvalues = [moving_average(v, mov_avg) for v in allvalues]
            timestamps = [moving_average(t, mov_avg) for t in timestamps]
        mean_rewards, std_dev_rewards, common_timesteps = interpolate_data(allvalues, timestamps)
        sns.lineplot(x=common_timesteps, y=mean_rewards, linestyle='-', linewidth=3, alpha=0.7,color=pallete[i],label=agent)
        plt.fill_between(common_timesteps, mean_rewards - std_dev_rewards, mean_rewards + std_dev_rewards, color=pallete[i], alpha=0.3)
    plt.xlabel('Time Step')
    plt.ylabel('Episodic reward')
    plt.legend()
    plt.show()


def interpolate_data(allvalues, timestamps):
    mins = []
    maxs = []
    for times in timestamps:
        mins.append(min(times))
        maxs.append(max(times))
    min_time = min(mins)
    max_time = max(maxs)-1000
    common_timesteps = np.linspace(min_time, max_time, 1000)
    interpolated_rewards = []
    for i in range(len(allvalues)):
        interp_func = interpolate.interp1d(timestamps[i], allvalues[i], kind='linear', bounds_error=False, fill_value='extrapolate')
        interpolated_rewards.append(interp_func(common_timesteps))
    interpolated_rewards = np.array(interpolated_rewards)

    # Calculate mean and standard deviation
    mean_rewards = np.mean(interpolated_rewards, axis=0)
    std_dev_rewards = np.std(interpolated_rewards, axis=0)
    return mean_rewards, std_dev_rewards, common_timesteps


def get_goal_values_for_successful_evaluations(experiment_name: str, db_url: str = None, oldsparse=False) -> List[float]:
    """
    Retrieves all goal values from successful evaluations for a given experiment name.

    :param experiment_name: The name of the experiment.
    :param db_url: Optional; The database URL. If None, will be obtained from config.
    :return: A list of goal values for successful evaluations.
    """
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()

    if experiment:
        # Query for goal values where Evaluation.success is True for the specific experiment
        max_step_subquery = (
            session.query(Evaluation.goal_value, func.max(Evaluation.step).label("max_step"))
            .filter(Evaluation.experiment_id == experiment.experiment_id, Evaluation.success.is_(True))
            .group_by(Evaluation.goal_value)
            .subquery()
        )
        if oldsparse:
            successful_evaluations = (
                session.query(Evaluation.goal_value, Evaluation.reward)
                .join(max_step_subquery, and_(
                    Evaluation.goal_value == max_step_subquery.c.goal_value,
                    Evaluation.step == max_step_subquery.c.max_step,
                    Evaluation.experiment_id == experiment.experiment_id,
                ))
                .all()
            )
            successful_goal_values = [evaluation.goal_value for evaluation in successful_evaluations if evaluation.reward > 0]
            failed_goal_values = [evaluation.goal_value for evaluation in successful_evaluations if not evaluation.reward > 0]
        else:
            # Query for goal values with the largest step value for each goal
            successful_evaluations = (
                session.query(Evaluation.goal_value, Evaluation.success)
                .join(max_step_subquery, and_(
                    Evaluation.goal_value == max_step_subquery.c.goal_value,
                    Evaluation.step == max_step_subquery.c.max_step,
                    Evaluation.experiment_id == experiment.experiment_id,
                ))
                .all()
            )
            # Extract goal values
            successful_goal_values = [evaluation.goal_value for evaluation in successful_evaluations if evaluation.success]
            failed_goal_values = [evaluation.goal_value for evaluation in successful_evaluations if not evaluation.success]

    session.close()
    engine.dispose()
    def convert_to_int_list(data):
        result = []
        for item in data:
            # Remove the brackets and split by spaces
            item = item.strip('[]')
            elements = item.split()
            # Convert the elements to integers
            int_elements = [float(x) for x in elements]
            result.append(int_elements)
        return result

    return convert_to_int_list(successful_goal_values), convert_to_int_list(failed_goal_values)

def get_goal_values_for_evaluations_grouped_by_step(experiment_name: str, db_url: str = None, oldsparse=False) -> Dict[int, List[List[float]]]:
    """
    Retrieves all goal values from evaluations for a given experiment name, grouped by step.

    :param experiment_name: The name of the experiment.
    :param db_url: Optional; The database URL. If None, will be obtained from config.
    :return: A dictionary where keys are steps and values are lists of goal values for that step.
    """
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    evaluations_grouped_by_step = {}

    if experiment:
        # Query for all evaluations, grouping by step
        evaluations = (
            session.query(Evaluation.step, Evaluation.goal_value, Evaluation.success, Evaluation.reward)
            .filter(Evaluation.experiment_id == experiment.experiment_id)
            .order_by(Evaluation.step)
            .all()
        )

        for evaluation in evaluations:
            step = evaluation.step
            goal_value = evaluation.goal_value
            success = evaluation.success
            reward = evaluation.reward

            if oldsparse:
                if reward > 0:
                    evaluations_grouped_by_step.setdefault(step, [[],[]])[0].append(goal_value)
                else:
                    evaluations_grouped_by_step.setdefault(step, [[],[]])[1].append(goal_value)
            else:
                if success:
                    evaluations_grouped_by_step.setdefault(step, [[],[]])[0].append(goal_value)
                else:
                    evaluations_grouped_by_step.setdefault(step, [[],[]])[1].append(goal_value)

    session.close()
    engine.dispose()

    def convert_to_int_list(data):
        result = []
        for item in data:
            # Remove the brackets and split by spaces
            item = item.strip('[]')
            elements = item.split()
            # Convert the elements to integers
            int_elements = [float(x) for x in elements]
            result.append(int_elements)
        return result

    # Convert goal values to int lists and group by step
    evaluations_grouped_by_step_converted = {step: [convert_to_int_list(goal_values[0]), convert_to_int_list(goal_values[1])] for step, goal_values in evaluations_grouped_by_step.items()}

    return evaluations_grouped_by_step_converted


def plot_maze_with_results(maze, successful_goal_pos):
    # Convert the maze to a numpy array for easier manipulation
    maze = np.array(maze)

    # Create a dictionary to count the number of successful hits for each goal position
    hit_counts = {}
    for pos in successful_goal_pos:
        pos_tuple = tuple(pos)
        if pos_tuple in hit_counts:
            hit_counts[pos_tuple] += 1
        else:
            hit_counts[pos_tuple] = 1

    # Create a color map for the goal hits
    cmap = mcolors.ListedColormap(['white', 'lightgreen', 'green', 'darkgreen', 'darkslategrey'])
    norm = mcolors.BoundaryNorm([0, 1, 2, 3, 4, 5], cmap.N)

    # Create the plot
    fig, ax = plt.subplots()
    for y in range(maze.shape[0]):
        for x in range(maze.shape[1]):
            if (maze[y, x]) == '1':
                color = 'black'
            elif maze[y, x] == 'r':
                color = 'red'
            elif maze[y, x] == 'g':
                hits = hit_counts.get((y, x), 0)
                color = cmap(norm(hits))
            else:
                color = 'white'
            print(f"pos: {y, x}, color: {color}, maze[y, x]: {maze[y, x]}, hit_counts: {hit_counts.get((y, x), 0)}")
            rect = plt.Rectangle([x, y], 1, 1, facecolor=color)
            ax.add_patch(rect)
    
    # Set the ticks and labels
    ax.set_xticks(np.arange(maze.shape[1]) + 0.5)
    ax.set_yticks(np.arange(maze.shape[0]) + 0.5)
    ax.set_xticklabels(np.arange(maze.shape[1]))
    ax.set_yticklabels(np.arange(maze.shape[0]))
    ax.invert_yaxis()

    plt.grid(which='both')
    plt.show()

def plot_maze_with_results_for_each_step(maze, evaluations_grouped_by_step):
    # Convert the maze to a numpy array for easier manipulation
    maze = np.array(maze)

    # Create a color map for the goal hits
    cmap = mcolors.ListedColormap(['white', 'lightgreen', 'green', 'darkgreen', 'darkslategrey'])
    norm = mcolors.BoundaryNorm([0, 1, 2, 3, 4, 5], cmap.N)

    for step, goal_positions in evaluations_grouped_by_step.items():
        # Reset hit counts for each step
        hit_counts = {}
        for pos in goal_positions[0]:
            pos_tuple = tuple(pos)
            hit_counts[pos_tuple] = hit_counts.get(pos_tuple, 0) + 1

        # Create the plot for the current step
        fig, ax = plt.subplots()
        ax.set_title(f'Step: {step}')
        for y in range(maze.shape[0]):
            for x in range(maze.shape[1]):
                if maze[y, x] == '1':
                    color = 'black'
                elif maze[y, x] == 'r':
                    color = 'red'
                elif maze[y, x] == 'g':
                    hits = hit_counts.get((y, x), 0)
                    color = cmap(norm(hits))
                else:
                    color = 'white'
                rect = plt.Rectangle([x, y], 1, 1, facecolor=color)
                ax.add_patch(rect)

        # Set the ticks, labels, and grid
        ax.set_xticks(np.arange(maze.shape[1]) + 0.5)
        ax.set_yticks(np.arange(maze.shape[0]) + 0.5)
        ax.set_xticklabels(np.arange(maze.shape[1]))
        ax.set_yticklabels(np.arange(maze.shape[0]))
        ax.invert_yaxis()
        plt.grid(which='both')

        # Display the plot for the current step
        plt.show()


def get_all_goals_for_experiment(experiment_name):
    # Create a SQLAlchemy engine and session
    db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    tag = session.query(Tag).filter_by(name='train/goal').first()
    if experiment and tag:
        # Query for all goal values for the specific experiment from the data_points table
        goals = (
            session.query(DataPoint.string_value)
            .filter_by(experiment_id=experiment.experiment_id, tag_id=tag.tag_id)
            .all()
        )
        session.close()
        goals = [[float(value) for value in goal[0].replace('[', '').replace(']', '').split()] for goal in goals]
        return goals
    
def get_successful_goals_for_experiment(experiment_name):
    # Create a SQLAlchemy engine and session
    db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    # Get all successful evaluations for the experiment only for the most recent steps
    max_step = session.query(func.max(Evaluation.step)).filter(Evaluation.experiment_id == experiment.experiment_id).first()[0]
    successful_evaluations = (
        session.query(Evaluation.goal_value)
        .filter(Evaluation.experiment_id == experiment.experiment_id, Evaluation.step == max_step, Evaluation.success == True)
        .all()
    )
    session.close()

    goals = list(set(tuple([float(value) for value in goal[0].replace('[', '').replace(']', '').split()]) for goal in successful_evaluations))
    #only extract unique goals
    return goals
    
def get_all_goals_for_study(study_name, db_name = None):
    # Create a SQLAlchemy engine and session
    db_url = get_default_storage(db_name=db_name)
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the study ID using the study name
    study_id = session.query(StudyModel.study_id).filter(StudyModel.study_name == study_name).first()[0]
    experiments = session.query(Experiment.name).filter(Experiment.study_id == study_id).all()
    session.close()
    d = {}
    for experiment in experiments:
        goals = get_all_goals_for_experiment(experiment.name)
        # print(f"Goals for experiment {experiment.name}: {goals}")
        d[experiment.name] = goals
    return d

def get_all_data_points_for_experiment(experiment_name, db_url=None):
    """
    Retrieves all data points for a given experiment name.

    :param experiment_name: The name of the experiment.
    :param db_url: Optional; The database URL. If None, will be obtained from config.
    :return: A list of data points for the given experiment.
    """
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    data_points = []

    if experiment:
        # Query for all data points for the specific experiment
        data_points = (
            session.query(DataPoint)
            .filter(DataPoint.experiment_id == experiment.experiment_id)
            .all()
        )

    session.close()
    engine.dispose()

    return data_points

def get_all_evaluations_for_experiment(experiment_name, db_url=None):
    """
    Retrieves all evaluations for a given experiment name.

    :param experiment_name: The name of the experiment.
    :param db_url: Optional; The database URL. If None, will be obtained from config.
    :return: A list of evaluations for the given experiment.
    """
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    evaluations = []

    if experiment:
        # Query for all evaluations for the specific experiment
        evaluations = (
            session.query(Evaluation)
            .filter(Evaluation.experiment_id == experiment.experiment_id)
            .all()
        )

    session.close()
    engine.dispose()

    return evaluations

def get_experiments_with_suffix(suffix, db_url=None):
    """
    Retrieves all experiments with names ending with the given suffix.

    :param suffix: The suffix to search for in experiment names.
    :param db_url: Optional; The database URL. If None, will be obtained from config.
    :return: A list of experiments with names ending with the given suffix.
    """
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    # Query for experiments with names ending with the given suffix
    experiments = session.query(Experiment).filter(Experiment.name.like(f"{suffix}%")).all()

    session.close()
    engine.dispose()

    return experiments

def plot_histogram_of_tag(results:dict, tag_name:str, bins:Union[int,float]=10):

    # Assuming `results` is your dictionary from the previous task
    proxy_results = results.copy()
    for key, value in results.items():
        if not value["hyperparameters"][0].get(tag_name):
            proxy_results.pop(key)

    bin_width = 0.005
    # Extract the number of layers for each experiment with objective value > 0.5
    layer_counts = []
    for study_name, study_data in proxy_results.items():
        for objective_value, params in zip(study_data['objective_values'], study_data['hyperparameters']):
            net_arch_length = round(params.get(tag_name, 0) / bin_width) * bin_width
            if objective_value > 0.5:
                layer_counts.append(net_arch_length)

    # Create a DataFrame with the number of layers
    layer_counts_df = pd.DataFrame({'num_layers': layer_counts})

    # Define bins and x-ticks
    bins = np.arange(min(layer_counts_df['num_layers']), max(layer_counts_df['num_layers']) + bin_width, bin_width)
    x_ticks = np.arange(min(layer_counts_df['num_layers']), max(layer_counts_df['num_layers']) + bin_width, bin_width)

    # Plot the histogram
    plt.figure(figsize=(10, 6))
    plt.hist(layer_counts_df['num_layers'], bins=bins, alpha=0.7, edgecolor='black')
    plt.xlabel(f'Number of {tag_name} (Rounded to Nearest {bin_width})')
    plt.ylabel('Count of Experiments')
    plt.xticks(x_ticks, rotation=45)
    plt.grid(True)
    plt.show()

def save_trial_plot(study_name, file_path):
    file_path = convert_fp_to_pathlib_create_directories(file_path)
    sns.set_theme()
    sns.set_context('talk')
    sns.set_style('darkgrid', {"axes.facecolor": "0.9", "plot.facecolor": "0.9"})

    plt.rcParams['axes.titleweight'] = 'bold'
    plt.rcParams['figure.facecolor'] = 'lightgrey'
    plt.rcParams['text.usetex'] = True
    plt.rcParams['xtick.labelsize'] = 18  # Increase tick label size
    plt.rcParams['ytick.labelsize'] = 12  # Increase tick label size
    plt.rcParams['axes.labelsize'] = 18  # Increase axis label size
    plt.rcParams['axes.labelweight'] = 'bold'  # Increase axis 

    PALETTE_STRING = 'rocket'
    palette = sns.color_palette(PALETTE_STRING, 3)
    storage_url = get_default_storage()
    study = optuna.load_study(study_name=study_name, storage=storage_url)
    
    # Extract trial numbers and objective values
    trials = study.trials_dataframe()
    trials = trials[trials['state'] == 'COMPLETE']
    trial_numbers = trials['number']
    objective_values = trials['value']

    # Calculate the cumulative maximum objective value
    cumulative_max = objective_values.cummax()
    cumulative_average = objective_values.expanding().mean()


    # Plot using Seaborn
    plt.figure(figsize=(10, 6))
    sns.scatterplot(x=trial_numbers, y=objective_values, color=palette[0], alpha=0.7, label='Trial')
    sns.lineplot(x=trial_numbers, y=cumulative_max, color = palette[1], alpha=0.7, label='Cumulative Max')
    sns.lineplot(x=trial_numbers, y=cumulative_average, color=palette[2], label='Cumulative Average')

    plt.xlabel('Trial Number')
    plt.ylabel(r'$\mathbf{f_{objective}}$')
    # plt.title('Trial Number vs. Objective Value')
    plt.legend()
    # plt.legend().remove()
    plt.tight_layout()
    # # Save the figure to file
    plt.savefig(file_path / f'{study_name}_trials.jpg', dpi=600)
    # # Close the plot to suppress automatic display
    plt.close()
    # plt.show()

def save_importance_plot(study_name, file_path, plotting_dict):
    file_path = convert_fp_to_pathlib_create_directories(file_path)
    sns.set_theme()
    sns.set_context('talk')
    sns.set_style('darkgrid', {"axes.facecolor": "0.9", "plot.facecolor": "0.9"})

    plt.rcParams['axes.titleweight'] = 'bold'
    plt.rcParams['figure.facecolor'] = 'lightgrey'
    plt.rcParams['text.usetex'] = True
    plt.rcParams['xtick.labelsize'] = 18  # Increase tick label size
    plt.rcParams['ytick.labelsize'] = 12  # Increase tick label size
    plt.rcParams['axes.labelsize'] = 18  # Increase axis label size
    plt.rcParams['axes.labelweight'] = 'bold'  # Increase axis 

    PALETTE_STRING = 'rocket'

    storage_url = get_default_storage()
    study = optuna.load_study(study_name=study_name, storage=storage_url)
    # Extract hyperparameter importance
    importance = optuna.importance.get_param_importances(study)

    # Convert the importance dictionary to a list of tuples and sort it
    importance = sorted(importance.items(), key=lambda x: x[1], reverse=True)

    # Separate the parameter names and their importance values
    params, values = zip(*importance)
    plot_names = [plotting_dict[param][1] for param in params]
    # Plot using Seaborn
    sns.barplot(x=values, y=plot_names, hue=plot_names, palette=PALETTE_STRING, legend=False)
    plt.xlabel('Importance')
    plt.ylabel('Hyperparameter')
    plt.title('Hyperparameter Importance')
    plt.tight_layout()
    # Save the figure to file
    plt.savefig(file_path + f'\\{study_name}_importance.jpg', dpi=600)
    # Close the plot to suppress automatic display
    plt.close()

def plot_save_SHAP(results, plotting_dict, combinations, file_path, experiments):
    file_path = os.path.join(file_path, '-'.join(experiments))
    file_path = convert_fp_to_pathlib_create_directories(file_path)

    X = []
    Y = []
    for study_name, study_data in results.items():
        if study_name in experiments:
            for objective_value, params in zip(study_data['objective_values'], study_data['hyperparameters']):
                X.append(params)
                Y.append(objective_value)

    # Convert X and Y to DataFrames
    X = pd.DataFrame(X)
    # change the column names
    for column in X.columns:
        X.rename(columns={column: plotting_dict[column][1]}, inplace=True)
    Y = pd.DataFrame(Y, columns=['objective_value'])
    X.fillna(0, inplace=True)
    print(X.shape)
    print(Y.shape)

    random_state = 42
    # Train/test split
    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=0.2, random_state=42)

    # Train a Random Forest model
    model = RandomForestRegressor(random_state=42)
    model.fit(X_train, Y_train)

        # Predict on the test set
    Y_pred = model.predict(X_test)

    # Evaluate the model
    mse = mean_squared_error(Y_test, Y_pred)
    r2 = r2_score(Y_test, Y_pred)
    print(f"Mean Squared Error for {'-'.join(experiments)}: {mse}")
    print(f"R^2 Score for {'-'.join(experiments)}: {r2}")

    cmap = sns.color_palette(PALETTE_STRING, as_cmap=True)
    # palette = sns.color_palette(PALETTE_STRING, as_cmap=True)

    # Create a custom colormap by trimming the top of the range
    # cmap = LinearSegmentedColormap.from_list('trimmed_palette', palette(np.linspace(0, 0.8, 256)))

    sns.set_theme()
    sns.set_context('talk')
    sns.set_style('darkgrid', {"axes.facecolor": "0.6", 
                               "plot.facecolor": "0.6", 
                               "grid.color": "0.5", 
                               "axes.edgecolor": "0.5"})

    plt.rcParams['axes.titleweight'] = 'bold'
    plt.rcParams['figure.facecolor'] = 'lightgrey'
    plt.rcParams['text.usetex'] = True
    plt.rcParams['xtick.labelsize'] = 20  # Increase tick label size
    plt.rcParams['ytick.labelsize'] = 20  # Increase tick label size
    plt.rcParams['axes.labelsize'] = 20  # Increase axis label size
    plt.rcParams['axes.labelweight'] = 'bold'  # Increase axis label weight
    # Initialize SHAP explainer
    explainer = shap.Explainer(model, X)
    shap_values = explainer(X)

    # SHAP summary plot
    shap.summary_plot(shap_values, X, cmap=cmap, show=False)
    plt.gcf().tight_layout()
    plt.gcf().savefig(os.path.join(file_path, "summary.png"), dpi=600)
    plt.close()
    
    for combo in combinations:
        # SHAP dependence plot
        shap.dependence_plot(combo[0], shap_values.values, X, cmap=cmap, interaction_index=combo[1], show=False)
        plt.tight_layout()
        c1 = combo[0].replace('\\','').replace('$','').replace('mathbf{','').replace('}','').replace('{','')
        c2 = combo[1].replace('\\','').replace('$','').replace('mathbf{','').replace('}','').replace('{','')
        plt.savefig(os.path.join(file_path, f"{c1}_{c2}.png"), dpi=600)
        plt.close()

def plot_and_save_histograms(results: dict, file_path: Union[str, pathlib.Path], bin_width: Union[int, float] = 10,
                             objective_value_threshold: float = 0.8, matching_string: str = 'hidden-layers',
                             legend_shift_factor: int = 1):
    set_sns_plot_settings()
    # Ensure the file path exists
    file_path = convert_fp_to_pathlib_create_directories(file_path)

    hidden_layer_counts = {}

    for study_name, study_data in results.items():
        for objective_value, params in zip(study_data['objective_values'], study_data['hyperparameters']):
            for key, val in params.items():
                if val is None or "length" in key:
                    continue
                if matching_string in key:
                    if key not in hidden_layer_counts:
                        hidden_layer_counts[key] = []
                    net_arch_length = round(val / bin_width) * bin_width
                    if objective_value > objective_value_threshold:
                        hidden_layer_counts[key].append(net_arch_length)

    layers = len(list(hidden_layer_counts.keys()))
    num = 0
    # Plot hidden layers
    if len(hidden_layer_counts) > 0:
        fig, ax = plt.subplots()
        for key, counts in hidden_layer_counts.items():
            if len(counts) > 0:
                bin_offset = num * (bin_width//layers)
                counts_df = pd.DataFrame({'num_layers': counts})
                bins = np.arange(min(counts_df['num_layers']) , max(counts_df['num_layers']) + legend_shift_factor*bin_width, bin_width//layers) + bin_offset
                sns.histplot((counts_df['num_layers']+bin_offset), bins=bins, ax=ax, color=sns.color_palette(PALETTE_STRING)[num], kde=False, label=f"Layer {num}", edgecolor='gray', linewidth=1)
                num+=1
        ax.set_xlabel(r"Neurons in Each Layer(Rounded to Nearest {})".format(bin_width))
        ax.set_ylabel(r'$\mathbf{{f_{{objective}} > {}}}$'.format(objective_value_threshold))
        ax.legend()
        plt.tight_layout()
        the_str = matching_string.replace('-', '_')
        print(f'Saving to {os.path.join(file_path, f"{the_str}_histogram.jpg")}')

        # Save the figure
        fig.savefig(os.path.join(file_path, f"{the_str}_histogram.jpg"), dpi=600)
        plt.close(fig)

def plot_heatmap_on_maze(maze, counts):
    """
    Plots a heatmap of counts on the maze.

    :param maze: A 2D list representing the maze with walls (1), paths (0), robot start ('r'), and goal ('g').
    :param counts: A Counter object where keys are locations (tuples) and values are counts.
    """
    # Convert the maze to a numpy array with numeric values for easier manipulation
    maze_numeric = []
    for row in maze:
        maze_numeric.append([1 if cell == 1 else 2 if cell == 'r' else 3 if cell == 'g' else 0 for cell in row])
    maze_numeric = np.array(maze_numeric, dtype=int)

    # Create a continuous color map for the counts
    heatmap_cmap = plt.cm.viridis
    heatmap_norm = mcolors.Normalize(vmin=0, vmax=max(counts.values()))

    # Create the plot
    fig, ax = plt.subplots()
    ax.set_title('Heatmap of Counts on Maze')

    # Plot the maze with text annotations
    for y in range(maze_numeric.shape[0]):
        for x in range(maze_numeric.shape[1]):
            if maze[y][x] == 1:
                ax.text(x, y, 'W', ha='center', va='center', color='black')
            elif maze[y][x] == 'r':
                ax.text(x, y, 'S', ha='center', va='center', color='red')
            elif maze[y][x] == 'g':
                ax.text(x, y, 'G', ha='center', va='center', color='blue')

    # Create a heatmap array with the same shape as the maze
    heatmap = np.zeros_like(maze_numeric, dtype=float)

    # Fill the heatmap array with the counts from the counter
    for (i, j), count in counts.items():
        heatmap[i, j] = count

    # Overlay the heatmap with transparency
    heatmap_img = ax.imshow(heatmap, cmap=heatmap_cmap, norm=heatmap_norm, interpolation='none', alpha=0.6)

    # Add a color bar to describe the heatmap
    cbar = plt.colorbar(heatmap_img, ax=ax, fraction=0.046, pad=0.04)
    cbar.set_label('Counts')

    # Set the ticks, labels, and grid
    ax.set_xticks(np.arange(maze_numeric.shape[1]) + 0.5)
    ax.set_yticks(np.arange(maze_numeric.shape[0]) + 0.5)
    ax.set_xticklabels(np.arange(maze_numeric.shape[1]))
    ax.set_yticklabels(np.arange(maze_numeric.shape[0]))
    ax.invert_yaxis()
    plt.grid(which='both')

    # Display the plot
    plt.show()

def get_N_best_experiments_in_study(study_name: str, N: int, 
                                    excluded_studies:List[str]=[], 
                                    db_url: str = None) -> List[str]:
    exps = get_best_experiments_in_study(study_name, 0.0, db_url=db_url)[::-1]
    experis = []
    for i in range(len(exps)):
        for j in range(len(exps[i])):
            if len(experis)>=N:
                return experis
            if exps[i][j][-1] in excluded_studies:
                continue
            experis.append(exps[i][j][-1])
    return experis
    
def plot_experiments_for_tags(experis:str, tag_comps:List[Tuple[str,str, str]], file_path: Union[str, pathlib.Path], exp_rename:str=''):
    
    for tag, comp, y_label in tag_comps:
        experiments, timestamps, allvalues = get_data_for_experiments_tag_and_component(experis[:], tag, comp)
        if exp_rename:
            for i in range(len(experiments)):
                experiments[i] = f"{exp_rename}_{i}"
        plot_data(experiments, 'Steps', timestamps, y_label, allvalues, xlims=[min(timestamps[0]), max(timestamps[0])], file_path=file_path)

def plot_coverage_for_experiments(experis:str, file_path: Union[str, pathlib.Path], exp_rename:str=''):
    e,s,sr = plot_evaluations_for_experiment(experis[:])
    if exp_rename:
        for i in range(len(e)):
            e[i] = f"{exp_rename}_{i}"
    plot_data(e, 'Steps', s, 'Coverage of Goal Space', sr, legend=True, file_path=file_path)

def plot_N_best_experiments_ep_rew_coverage(study_name: str, N: int, 
                                            file_path: Union[str, pathlib.Path], 
                                            db_url: str = None, exp_rename:str='', 
                                            excluded_studies:List[str]=[]):
    file_path = convert_fp_to_pathlib_create_directories(file_path)
    experis = get_N_best_experiments_in_study(study_name, N, 
                                              excluded_studies=excluded_studies, 
                                              db_url=db_url)
    if not experis:
        return
    plot_exp_rew_coverage(experis, file_path, study_name, exp_rename=exp_rename)

def plot_exp_rew_coverage(experis: List[str], file_path: Union[str, pathlib.Path], file_name:str, exp_rename:str=''):
    file_path = convert_fp_to_pathlib_create_directories(file_path)
    plot_experiments_for_tags(experis, [('train/ep_rew', 'SACSB3Agent', 'Episodic Reward')], file_path/f'{file_name}_ep_rew.png', exp_rename=exp_rename)
    plot_coverage_for_experiments(experis, file_path / f'{file_name}_coverage.png', exp_rename=exp_rename)

def convert_fp_to_pathlib_create_directories(file_path: Union[str, pathlib.Path]):
    #check if filepath is string and convert to path
    is_dir = False
    if isinstance(file_path, str):
        is_dir = file_path[-1] == '/'
        file_path = pathlib.Path(file_path)
    #check if path is a directory and if it exists and create if not using pathlib if not directory create the directory above the file
    if not file_path.is_dir() and not is_dir:
        if not file_path.parent.exists():
            file_path.parent.mkdir(parents=True)
    else:
        if not file_path.exists():
            file_path.mkdir(parents=True)
    return file_path

def set_sns_plot_settings(label_size:int=20, x_tick_size:int=18, y_tick_size:int=18):
    sns.set_theme()
    sns.set_context('talk')
    sns.set_style('darkgrid', {"axes.facecolor": "0.9", "plot.facecolor": "0.9"})

    plt.rcParams['axes.titleweight'] = 'bold'
    plt.rcParams['figure.facecolor'] = 'lightgrey'
    plt.rcParams['text.usetex'] = True
    plt.rcParams['xtick.labelsize'] = x_tick_size  # Increase tick label size
    plt.rcParams['ytick.labelsize'] = y_tick_size  # Increase tick label size
    plt.rcParams['axes.labelsize'] = label_size  # Increase axis label size
    plt.rcParams['axes.labelweight'] = 'bold'  # Increase axis label weight


def get_all_metadata_for_experiment(experiment:str) -> Dict[str,str]:
    db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    experiment = session.query(Experiment).filter(Experiment.name == experiment).first()
    metadatas = {}
    if experiment:
       metadata = session.query(ExpMetadata).filter_by(experiment_id=experiment.experiment_id).all()
       for data in metadata:
           metadatas[data.key] = data.value
    session.close()
    engine.dispose()
    return metadatas




from abc import ABC, abstractmethod, abstractclassmethod
import pathlib
import argparse
import time

from ..baseclasses import ComponentCollection
from ..utility import check_create, save_json
from ..metaclasses.hyperparameter import HyperParser
from ...logger import DatabaseLogger


class ExperimentParser(HyperParser):
    _register = {}


class AbstractExperimentParser(type(ABC), ExperimentParser):
    pass


class AbstractExperiment(ABC, metaclass=AbstractExperimentParser):
    """Abstract experiment class for running core experiments

    Provides an interface as well as loading and saving methods.
    Will generate an experiment folder and add a json config for 
    all serialisable objects in the experiment.
    """
    _builders = []
    _tester = None
    _load = False

    def __init__(self, file_path: pathlib.Path,
                 exp_args: dict, *args, load = False, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)
        
        self._exp_args = exp_args
        self._file_path = file_path
        self._exp_args['file_path'] = str(file_path.parent)
        self._json = self._file_path / 'config.json'
        self._results = self._file_path / 'results.json'
        if not load:
            self._save()
        # self._component_path = self._file_path / 'components'
        # check_create(self._component_path)
        if DatabaseLogger._instance is None:
            self._logger = DatabaseLogger(str(self._file_path.stem), exp_metadata_dict=exp_args)
        else:
            self._logger = DatabaseLogger._instance
        self._components = ComponentCollection()
        self._trainer = None
        self.build()
        self._reorder_components()
        # self._tester = self._tester(self).build()
        if load:
            self._load()

    def run(self):
        self._components.run()
        logger_timeout = 0
        logger_qsize = 0
        while not self._logger.queue.empty():
            # for res in self._logger.results:
            #     res.get()
            new_logger_qsize = self._logger.queue.qsize()
            if new_logger_qsize != logger_qsize:
                logger_timeout = 0
                logger_qsize = new_logger_qsize
            else:
                logger_timeout += 1
            print(f"Waiting for logger to finish...queue size: {logger_qsize}")
            
            if logger_timeout > 10:
                break
            time.sleep(1)
        print("Closing Logger...")      
        self._logger.close()

    def test(self):
        if self._tester is None:
            return
        scalar, results = self._tester.test()

        results_dict = {
            'scalar': scalar,
            'results': list(results),
        }
        if self._tester._tester._initial_states is not None:
            initials = self._tester._tester._initial_states.tolist()
            results_dict['initial_states'] = initials
        if hasattr(self._tester, '_goals'):
            results_dict['goals'] = self._tester._goals._goal_schedule.tolist()
        save_json(self._results, results_dict)

    def build(self):
        temp_builders = self._builders[:]
        for i, builder in enumerate(temp_builders):
            self._builders[i] = builder(self)

    def _save(self):
        check_create(self._json)
        save_json(self._json, self._exp_args)

    def _reorder_components(self):
        pass

    @classmethod
    def harvest_parameters(cls):
        cls._parser = cls._arg_parser()
        cls.experiment_params()

    @classmethod
    def _arg_parser(cls):
        parents = [
            klass._parser for builder in cls._builders
            for klass in builder._register if hasattr(klass, "_parser")
        ]
        return argparse.ArgumentParser(parents=parents, add_help=None)

    @classmethod
    def experiment_params(cls):
        #Adds class experiment specific arguments to the experiment
        cls._parser.add_argument('--db-url', default=None, type=str, help='Database URL or Config file')

    def _load(self):
        for component in self._components:
            component.load()

    @property
    def components(self):
        return self._components



import sys
import os
import optuna
import argparse
import pathlib
import threading
import time
from typing import Dict, List, Tuple

from ..optimiser.hpoptimiser import OptunaObjective
from ..logger import get_default_storage
from ..interfaces.abstractbaseclasses import AbstractExperiment
from ..experimentrunners import LocalRunner, SBatchGenerator, FilePathGenerator, PexpectSFTPClient, RemoteSBatchSFTPRunner

def get_opt_inps(args:Dict[str, str], 
                 unknowns: List[str]) -> Tuple[str, str, Dict[str, str], str]:
    """Get the inputs for the optimisation from the command line arguments

    :param args: The known arguments
    :type args: Dict[str, str]
    :param unknowns: The arguments related to the experiment strategy
    :type unknowns: List[str]
    :return: The strategy, study name, optimisation kwargs and storage string
    :rtype: Tuple[str, str, Dict[str, str], str]
    """ 
    strategy = args.pop('strategy')
    study_name = args.pop('study_name')
    if study_name == None:
        study_name = strategy
    parser = argparse.ArgumentParser(parents=[AbstractExperiment._register[strategy]._parser])
    storage_string = get_default_storage() if args['storage'] is None else args['storage']
    args.pop('storage')
    if args['name'] == '':
        args.pop('name')
    parsed_unknowns = vars(parser.parse_args(unknowns))
    unknowns = [val.replace('--','').replace('-','_') for val in unknowns]
    non_default_args = {k: v for k, v in parsed_unknowns.items() if k in unknowns}
    opt_kwargs = {**args, **non_default_args}
    return strategy, study_name, opt_kwargs, storage_string

def get_storage(storage_string:str)->optuna.storages.RDBStorage:
    """Get the storage object for the experiment

    :param storage_string: the SQLAlchemy RDB String
    :type storage_string: str
    :return: The RDBStorage object
    :rtype: optuna.storages.RDBStorage
    """   
    storage = optuna.storages.RDBStorage(
                url=storage_string,
                engine_kwargs={'pool_recycle': 3600, 'pool_pre_ping': True}
            )
    return storage

def create_and_run_remote_study(strategy: str, 
                                study_name: str, 
                                opt_kwargs: Dict[str,str], 
                                storage: optuna.storages.RDBStorage,
                                remote_prompt: str,
                                remote_path: str,
                                sbatch_config: str,
                                config_kwargs: Dict[str,str] = None,
                                ssh_name: str = None,
                                file_path: str = None,
                                storage_string: str = None
                                ) -> None:
    """Create and run a remote study

    :param strategy: The experiment
    :type strategy: str
    :param study_name: The name of the optuna study
    :type study_name: str
    :param opt_kwargs: The static optimiser kwargs
    :type opt_kwargs: Dict[str,str]
    :param storage: The storage object
    :type storage: optuna.storages.RDBStorage
    :param remote: the name of the remote host
    :type remote: str
    :param remote_prompt: the prompt on your remote
    :type remote_prompt: str
    :param remote_path: the location to store experiment files on the remote host
    :type remote_path: str
    """   
    sampler = opt_kwargs.pop('sampler', 'tpe')
    study = optuna.create_study(storage = storage,
                                    study_name = study_name, 
                                    direction='maximize', 
                                    load_if_exists=True,
                                    sampler = create_sampler(sampler))
    file_path = file_path / 'SB3ExperimentsBatchFiles' / study.study_name
    generator = SBatchGenerator(study.study_name, sbatch_config, config_kwargs)
    sftp = PexpectSFTPClient(ssh_name, remote_prompt)
    file_path_generator = FilePathGenerator(file_path, remote_path)
    runner = RemoteSBatchSFTPRunner(file_path, sftp, generator, file_path_generator)
    objective = OptunaObjective(AbstractExperiment._register[strategy], runner, opt_kwargs=opt_kwargs, storage=storage_string)
    study.optimize(objective.objective, n_trials=1, n_jobs=1)

def create_sampler(sampler:str):
    if sampler == 'tpe':
        print("Using TPE Sampler")
        return optuna.samplers.TPESampler()
    elif sampler == 'gp':
        print("Using GP Sampler")
        return optuna.samplers.GPSampler(deterministic_objective=True)
    elif sampler == 'cmaes':
        print("Using CMAES Sampler")
        return optuna.samplers.CmaEsSampler()
    else:
        raise ValueError("Invalid sampler")

def run_local(args: Dict[str, str], unknowns: List[str], runner_script: str):
    """Run the experiments locally

    :param args: the known arguments
    :type args: Dict[str, str]
    :param unknowns: the experiment arguments
    :type unknowns: List[str]
    """
    N_JOBS = args.pop('num_jobs')
    sampler = args.pop('sampler', 'tpe') 
    assert runner_script is not None  
    strategy , study_name, opt_kwargs, storage_string = get_opt_inps(args, unknowns)
    
    for _ in range(N_JOBS):
        storage = get_storage(storage_string)
        study = optuna.create_study(storage = storage,
                                    study_name = study_name, 
                                    direction='maximize', 
                                    load_if_exists=True,
                                    sampler = create_sampler(sampler))
        runner = LocalRunner(runner_script, storage_string)
        objective = OptunaObjective(AbstractExperiment._register[strategy], runner, opt_kwargs=opt_kwargs, storage=storage_string)
        study.optimize(objective.objective, n_trials=1, n_jobs=1)

def run_remote(args, unknowns, remote_prompt, remote_path, sbatch_config, config_kwargs, ssh_name):
    N_JOBS = args.pop('num_jobs')
    path = args.pop('path', None)
    assert path is not None
    strategy, study_name, opt_kwargs, storage_string = get_opt_inps(args, unknowns)
    for _ in range(N_JOBS):
        storage=get_storage(storage_string)
        create_and_run_remote_study(strategy, study_name, opt_kwargs, storage, remote_prompt, remote_path, sbatch_config, config_kwargs, ssh_name, file_path=pathlib.Path(path), storage_string=storage_string)

def runopt(args, unknowns, remote_prompt:str=None, remote_path:str=None, sbatch_config:str=None, config_kwargs:Dict[str,str]=None, ssh_name:str=None):
    computer = args.pop('computer', None)
    assert computer is not None
    if computer == "remote":
        assert remote_prompt is not None
        assert remote_path is not None
        assert ssh_name is not None
    if computer == "local":
        runner_script = args.pop('runner_script', None)
        assert runner_script is not None
    N_RUNS = args.pop('num_runs', 5)
    threads = []    
    for _ in range(N_RUNS):
        if computer == 'remote':
            thread = threading.Thread(target=run_remote, args=(args.copy(), unknowns.copy(), remote_prompt, remote_path, sbatch_config, config_kwargs, ssh_name))
            thread.start()
            threads.append(thread)
        if computer == 'local':
            thread = threading.Thread(target=run_local, args=(args.copy(), unknowns.copy(), runner_script))
            thread.start()
            threads.append(thread)
        time.sleep(30)
    for thread in threads:
        thread.join()



from typing import List, Union, Dict, Callable, Tuple
from abc import abstractmethod
from threading import Thread, Event
from ...logger import DatabaseLogger

import time
import sys

from .hyperparameter import AbstractParametered
# from ..patterns import Observer, Subject
from .io import MasterIO

class AbstractComponent(AbstractParametered, Thread):
    """
    The basic core building block of an RLOS experiment. Components are isolated
    algorithmic sections that can interact with one another via Observer, Subject
    singletons. 
    
    A Component uses Subjects to output information and Observers as inputs.
    The use of this paradigm allows components to be agnostic of each other. All
    that is required is that every Subject has an Observer in the Experiment space. 

    Note: The inputs and outputs can be either a list or a dictionary. If they are a list the 
    attributes of the class and the names of the observers/subjects will be each element of the list. If they 
    are a dictionary then the keys will represent the name of the attribute in the class and the item 
    will represent the name of the observer/subject.
    """
    _is_training = True
    _all_instances_started = Event()
    _all_instances_finished = Event()
    _num_live_instances = 0
    _error_event = Event()

    def __init__(self, io_map: Dict[str,str], additional_io_methods:List[str]=[], *args, **kwargs):
        """
        Constructor of the Component base class

        :param inputs: A dict of list or dict of names that contain the input namespace 
        :type inputs: Dict[Union[Dict[str,str],List[str]]]
        :param outputs: A dict of list or dict of names that contain the output namespace
        :type outputs: Dict[Union[Dict[str,str],List[str]]]
        """
        super().__init__(*args, **kwargs)
        if DatabaseLogger._instance is not None:
            self._logger_inst = DatabaseLogger._instance.bind(f'{self.__class__.__name__}')
        else:
            self._logger_inst = None
        self._io_map = io_map
        self._additional_io_methods = additional_io_methods
        self.io = MasterIO(self)
        self._inputs = self.io.inputs
        self._outputs = self.io.outputs
        
        
    def _generate_sub_obs(self):
        [self.__setattr__(f'_{key}', val) for key,val in self.io.subjects.items()]
        [self.__setattr__(f'_{key}', val) for key, val in self.io.observers.items()]


    def _process_io(self):
        """
        Sets inputs/outputs as attributes of the class

        :param inp: Either the input or output of the component
        :type inp: Union[Dict[str,str],List[str]]
        :raises TypeError: input must be a dictionary or list
        """
        self._generate_sub_obs()
        
        #Getter for property
        def observer_getter_factory(name):
            def getter(self):
                # while getattr(self, f'_{name}').accessed:
                #     time.sleep(0.0005)
                return self.__dict__[f'_{name}'].state
            return getter
        
        #getter for subject property
        def getter_factory(name):
            def getter(self):
                return self.__dict__[f'_{name}'].state
            return getter
        
        #Setter for property
        def setter_factory(name):
            def setter(self, value):
                self.__dict__[f'_{name}'].state = value
            return setter
        
        # Create the private attributes and properties
        for k, _ in self.io.inputs.items():
            #Use getter factory as a property of the class
            setattr(self.__class__, k, property(observer_getter_factory(f'{k}'), None))
        for k, _ in self.io.outputs.items():
            setattr(self.__class__, k, property(getter_factory(f'{k}'), setter_factory(f'{k}')))
        
    def run(self):
        self._all_instances_started.wait()
        AbstractComponent._num_live_instances += 1
        # print(self._num_live_instances)
        try:
            self.do_run()
            self.done()
        except Exception as e:
            AbstractComponent._error_event.set()
            self.done()
            raise e
        self._all_instances_finished.wait()
        print(f'{self.__class__.__name__} finished')
    
    @abstractmethod
    def do_run(self):
        """
        Threads need to define run
        """        
        pass           

    @abstractmethod
    def save(self):
        """
        Save the component state so that an experiment can be resumed.
        """
        pass

    @abstractmethod
    def load(self):
        """
        Load the a saved component to resume correct state. 
        """
        pass

    def done(self):
        AbstractComponent._num_live_instances -= 1
        # print(self._num_live_instances)
        if self._num_live_instances == 0:
            # print('All instances finished')
            self._all_instances_finished.set()


from sqlalchemy import Column, Integer, String, UniqueConstraint, DateTime, Float, Boolean, ForeignKey, func
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

QUERY_BATCH = 200



class Study(Base):
    __tablename__ = "studies"
    study_id = Column(Integer, primary_key=True)
    study_name = Column(
        String, index=True, unique=True, nullable=False
    )

class Experiment(Base):
    __tablename__ = 'experiments'
    experiment_id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False, unique=True)
    start_time = Column(DateTime, nullable=False, default=func.now())
    end_time = Column(DateTime)
    study_id = Column(Integer, ForeignKey('studies.study_id', ondelete='CASCADE'))

class ExpMetadata(Base):
    __tablename__ = 'exp_metadata'
    exp_metadata_id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey('experiments.experiment_id', ondelete='CASCADE'), nullable=False)
    key = Column(String(255), nullable=False)
    value = Column(String(10240), nullable=False)

class Tag(Base):
    __tablename__ = 'tags'
    tag_id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False, unique=True)

class Component(Base):
    __tablename__ = 'components'
    component_id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False, unique=True)

class DataPoint(Base):
    __tablename__ = 'data_points'
    data_point_id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey('experiments.experiment_id', ondelete='CASCADE'), nullable=False)
    tag_id = Column(Integer, ForeignKey('tags.tag_id'), nullable=False)
    component_id = Column(Integer, ForeignKey('components.component_id'))
    int_value = Column(Integer)
    float_value = Column(Float)
    bool_value = Column(Boolean)
    string_value = Column(String(10240))
    step = Column(Integer)
    timestamp = Column(DateTime, nullable=False, default=func.now())
    __table_args__ = (
        UniqueConstraint('experiment_id', 'tag_id', 'component_id', 'step', name='uix_exp_tag_comp_id_step'),
    )

class Evaluation(Base):
    __tablename__ = 'evaluations'
    evaluation_id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey('experiments.experiment_id', ondelete='CASCADE'), nullable=False)
    init_position = Column(String(1020))
    goal_value = Column(String(510))
    reward = Column(Float)
    success = Column(Boolean)
    step = Column(Integer)



from multiprocessing import Queue, Event

from queue import Empty
import re
import time
import logging
from datetime import datetime
import psycopg2

from sqlalchemy import create_engine, and_, func
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import IntegrityError, OperationalError

from .tables import DataPoint, Evaluation, Tag, QUERY_BATCH, ExpMetadata, Experiment, Base, Component

# logging.basicConfig(level=logging.DEBUG)

class DBWorker:
    

    def __init__(self, queue: Queue, stop_signal: Event): # type: ignore
        self.db_url = None
        self.main_experiment_id = None
        self.backup_db_url = None
        self.backup_experiment_id = None

        self.queue = queue
        self.stop_signal = stop_signal
        self.timeout = False
        self.timeout_time = 0
        self.experiment_id = None
        #set the current experiment id
        self.main_tag_cache = {}
        self.backup_tag_cache = {}
        self.main_component_cache = {}
        self.backup_component_cache = {}

        self.main_datapoint_queries = []
        self.backup_datapoint_queries = []
        self.main_evaluation_queries = []
        self.backup_evaluation_queries = []
        self.main_component_names = []
        self.backup_component_names = []

    
    def _is_backup(self):
        return self.experiment_id == self.backup_experiment_id

    def load_tags(self):
        session = self.Session_main()
        tags = []
        try:
            tags = session.query(Tag).all()
            self.main_tag_cache = {'ids': {tag.tag_id: tag.name for tag in tags}, 'names': {tag.name: tag.tag_id for tag in tags}}
            session.close()
        except OperationalError:
            session.close()
        session = self.Session_backup()
        tags_bu = session.query(Tag).all()            
        self.backup_tag_cache = {'ids': {tag.tag_id: tag.name for tag in tags_bu}, 'names': {tag.name: tag.tag_id for tag in tags_bu}}
        for tag in tags:
            if tag.name not in self.backup_tag_cache['names']:
                t = Tag(name=tag.name)
                session.add(t)
        session.commit()
        tags_bu = session.query(Tag).all()
        self.backup_tag_cache = {'ids': {tag.tag_id: tag.name for tag in tags_bu}, 'names': {tag.name: tag.tag_id for tag in tags_bu}}      
        session.close()

    def load_components(self):
        session = self.Session_main()
        components = []
        try:
            components = session.query(Component).all()
            self.main_component_cache = {'ids': {component.component_id: component.name for component in components}, 'names': {component.name: component.component_id for component in components}}
            session.close()
        except OperationalError:
            session.close()
        session = self.Session_backup()
        components_bu = session.query(Component).all()
        self.backup_component_cache = {'ids': {component.component_id: component.name for component in components_bu}, 'names': {component.name: component.component_id for component in components_bu}}
        for component in components:
            if component.name not in self.backup_component_cache['names']:
                c = Component(name=component.name)
                session.add(c)
        session.commit()
        components_bu = session.query(Component).all()
        self.backup_component_cache = {'ids': {component.component_id: component.name for component in components_bu}, 'names': {component.name: component.component_id for component in components_bu}}
        session.close()

    def convert_datapoint_queries(self):
        if self._is_backup():
            for _ in self.main_datapoint_queries.copy():
                query = self.main_datapoint_queries.pop(0)
                query['experiment_id'] = self.backup_experiment_id
                tag_name = self.main_tag_cache['ids'][query['tag_id']]
                query['tag_id'] = self.backup_tag_cache['names'][tag_name]
                component_name = self.main_component_cache['ids'].get(query['component_id'])
                query['component_id'] = self.backup_component_cache['names'][component_name]
                self.backup_datapoint_queries.append(query)
        else:
            for _ in self.backup_datapoint_queries.copy():
                query = self.backup_datapoint_queries.pop(0)
                query['experiment_id'] = self.main_experiment_id
                tag_name = self.backup_tag_cache['ids'][query['tag_id']]
                query['tag_id'] = self.main_tag_cache['names'][tag_name]
                component_name = self.backup_component_cache['ids'].get(query['component_id'])
                query['component_id'] = self.main_component_cache['names'][component_name]
                self.main_datapoint_queries.append(query)

    def convert_evaluation_queries(self):
        if self._is_backup():
            for _ in self.main_evaluation_queries.copy():
                query = self.main_evaluation_queries.pop(0)
                query['experiment_id'] = self.backup_experiment_id
                self.backup_evaluation_queries.append(query)
        else:
            for _ in self.backup_evaluation_queries.copy():
                query = self.backup_evaluation_queries.pop(0)
                query['experiment_id'] = self.main_experiment_id
                self.main_evaluation_queries.append(query)

    def convert_queries(self):
        self.convert_datapoint_queries()
        self.convert_evaluation_queries()

    @property
    def tag_cache(self):
        return self.main_tag_cache if not self._is_backup() else self.backup_tag_cache
    
    @property
    def component_cache(self):
        return self.main_component_cache if not self._is_backup() else self.backup_component_cache
    
    @property
    def datapoint_queries(self):
        return self.main_datapoint_queries if not self._is_backup() else self.backup_datapoint_queries
    
    @property
    def evaluation_queries(self):
        return self.main_evaluation_queries if not self._is_backup() else self.backup_evaluation_queries

    @property
    def Session(self):
        # logging.debug(f"Attempting to get Session")
        if not self.check_timeout():
            try:
                # Try to connect to the main database
                connection = self._engine.connect()
                connection.close()
                self.timeout = False
                if self.experiment_id is None:
                    self.experiment_id = self.main_experiment_id
                if self._is_backup():
                    self.experiment_id = self.main_experiment_id
                    self.convert_queries()
                logging.debug(f"Got main experiment")
                return self.Session_main
            except OperationalError:
                # If connection fails, use the backup database
                self.timeout = True
                self.timeout_time = time.time()
                if not self._is_backup():
                    self.experiment_id = self.backup_experiment_id
                    self.convert_queries()
                logging.debug(f"Got backup experiment")
                return self.Session_backup
        else:
            return self.Session_backup
    
    def check_timeout(self):
        # logging.debug(f"Timeout: {self.timeout}")
        if self.timeout:
            delta = time.time() - self.timeout_time
            return delta < 120
        else: 
            return False
        
    def check_run(self):
        keep_running =  not (self.stop_signal.is_set() and self.queue.empty())
        if not keep_running:
            logging.debug(f"Queue Loop Finished because stop_signal:{self.stop_signal.is_set()} and queue empty:{self.queue.empty()}")
        return keep_running

    def run(self):
        logging.debug(f"Starting Debug Worker")
        #Assumes postgresql
        if 'sqlite' in self.db_url:
            logging.debug(f"Creating sqlite engine")
            self._engine = create_engine(self.db_url)
        else:
            logging.debug(f"Creating postgresql engine")
            self._engine = create_engine(self.db_url, pool_pre_ping=True, connect_args={'connect_timeout': 10})
        logging.debug(f"Creating backup engine")
        self._engine_backup = create_engine(self.backup_db_url, pool_pre_ping=True)
        logging.debug(f"Creating session main")
        self.Session_main = sessionmaker(bind=self._engine)
        logging.debug(f"Creating session backup")
        self.Session_backup = sessionmaker(bind=self._engine_backup)
        logging.debug(f"Creating tags")
        self.load_tags()
        logging.debug(f"Creating components")
        self.load_components()
        logging.debug(f"Starting main loop")
        while self.check_run():
            #protect against empty self.queue
            logging.debug(f"Checking if queue is empty")
            if self.queue.empty():
                logging.debug(f"queue is empty")
                time.sleep(2)
            else:
                #two processes means maybe we do get empty self.queue so don't block
                try:
                    logging.debug(f"pulling from queue")
                    obj, method = self.queue.get(False)
                except Empty:
                    continue
                try:
                    if method == 'add_data_point':
                        try:
                            tag_id = self.get_tag_id(obj)
                            component_id = self.get_component_idz(obj)
                            # Create data point with tag ID
                            _ = self.Session
                            data_point = dict(experiment_id=self.experiment_id, tag_id=tag_id, component_id=component_id, step=obj.step,
                                                int_value=obj.int_value, float_value=obj.float_value,
                                                bool_value=obj.bool_value, string_value=obj.string_value,
                                                timestamp=obj.timestamp)
                            self.datapoint_queries.append(data_point)
                            if self._is_backup():
                                self.backup_component_names.append(obj.component_name)
                            else:
                                self.main_component_names.append(obj.component_name)
                        except Exception as e:
                            logging.debug(f"Failed to add obj: {obj} with method: {method} with error: {e}")
                            continue
                    elif method == "add_experiment_result":
                        try:
                            _ = self.Session
                            evaluation = dict(
                                experiment_id=self.experiment_id,
                                init_position=obj.init_position,
                                goal_value=obj.goal_value,
                                reward=obj.reward,
                                success=obj.success,
                                step=obj.step
                            )
                            self.evaluation_queries.append(evaluation)
                        except Exception as e:
                            logging.debug(f"Failed to add obj: {obj} with method: {method} with error: {e}")
                    else:
                        raise ValueError(f"Invalid method {method}")
                    logging.debug(f"Checking to add datapoints")
                    if len(self.datapoint_queries) > QUERY_BATCH:
                        try:
                            self.insert_bulk_data_points(self.datapoint_queries)
                        except Exception as e:
                            logging.debug(f"Failed to add obj: {obj} with method: {method} with error: {e}")
                    logging.debug(f"Checking to add evaluations")
                    if len(self.evaluation_queries) > 0:
                        try:
                            self.insert_bulk_evaluations(self.evaluation_queries)
                        except Exception as e:
                            logging.debug(f"Failed to add obj: {obj} with method: {method} with error: {e}")
                except Exception as e:
                    logging.debug(f"Failed to add obj: {obj} with method: {method} with error: {e}")
        logging.debug(f"Exiting Debug Worker")
        if len(self.datapoint_queries):
            self.insert_bulk_data_points(self.datapoint_queries)
        if len(self.evaluation_queries):
            self.insert_bulk_evaluations(self.evaluation_queries)
        

    
    def insert_bulk_evaluations(self, evaluation_queries):
        session = self.Session()
        session.bulk_insert_mappings(Evaluation, evaluation_queries)
        session.commit()
        session.close()
        evaluation_queries.clear()

    def insert_bulk_data_points(self, datapoint_queries):
        try:
            session = self.Session()
            session.bulk_insert_mappings(DataPoint, datapoint_queries)
            session.commit()
        except IntegrityError as e:
            session.rollback()
            if 'postgresql' in session.bind.engine.url.__str__():
                constraint_name = e.orig.diag.constraint_name
                if constraint_name == 'uix_exp_tag_comp_id_step':
                    pattern = r'\((.*?)\)'
                    matches = re.finditer(pattern, e.orig.diag.message_detail)
                    counter = 0
                    for match in matches:
                        if counter%2:
                            values = match.group(1).split(', ')
                            offending_data = dict(zip(keys, values))
                            logging.debug(f"REMOVING Offending data: {offending_data}")
                            session.query(DataPoint).filter(
                                and_(
                                    DataPoint.experiment_id == offending_data['experiment_id'],
                                    DataPoint.tag_id == offending_data['tag_id'],
                                    DataPoint.component_id == offending_data['component_id'],
                                    DataPoint.step == offending_data['step']
                                )
                            ).delete()
                            session.commit()
                        else:
                            keys = match.group(1).split(', ')
                        counter += 1
                    segments = self.split_list_into_four(datapoint_queries)
                    for segment in segments:
                        self.insert_bulk_data_points(segment)
            elif 'sqlite' in session.bind.engine.url.__str__():
                duplicates = session.query(DataPoint).group_by(
                    DataPoint.experiment_id, DataPoint.tag_id, DataPoint.component_id, DataPoint.step
                ).having(func.count('*') > 1).all()
                for duplicate in duplicates:
                    session.delete(duplicate)
                session.bulk_insert_mappings(DataPoint, datapoint_queries)
                session.commit()
        except OperationalError as e:
            session.close()
            time.sleep(5)
            segments = self.split_list_into_four(datapoint_queries)
            for segment in segments:
                self.insert_bulk_data_points(segment)
        except Exception as e:
            logging.debug(f"Failed to insert datapoints with error: {e}")
            raise e
        finally:
            session.close()
        datapoint_queries.clear()

    def split_list_into_four(self, lst):
        # Calculate the length of each segment
        length = len(lst)
        segment_size = length // 4
        extra = length % 4

        # Create the segments
        segments = []
        for i in range(4):
            start = i * segment_size + min(i, extra)
            end = start + segment_size + (1 if i < extra else 0)
            segments.append(lst[start:end])

        return segments

    def get_component_idz(self, obj):
        if not self._is_backup():
            return obj.component_id
        else:
            component_name = self.main_component_cache['ids'][obj.component_id]
            return self.backup_component_cache['names'][component_name]
        
    def get_tag_id(self, obj):
        # Check if tag ID is in the cache
        tag_id = self.tag_cache['names'].get(obj.tag_name, None)    
        # If not, check if tag with given name exists in the database
        session = self.Session()
        # If not, create it
        if tag_id is None:
            tag = Tag(name=obj.tag_name)
            try:
                session.add(tag)
                session.commit()
                if not self._is_backup():
                    bu_session = self.Session_backup()
                    bu_session.add(tag)
                    bu_session.commit()
                    bu_session.close()
            except Exception as e:
                session.rollback()
                time.sleep(5.0)
                tag = session.query(Tag).filter_by(name=obj.tag_name).first()
                if tag is None:
                    raise e  
            # Add tag name and ID to the cache
            tag_id = tag.tag_id
            self.tag_cache['names'][obj.tag_name] = tag_id
            session.close()
        return tag_id
    
    def add_exp_metadata(self, metadata_dict):
        self._add_exp_metadata(self.db_url, self.experiment_id, metadata_dict)
        if self.db_url != self.backup_db_url:
            self._add_exp_metadata(self.backup_db_url, self.backup_experiment_id, metadata_dict)
    
    def _add_exp_metadata(self, db_url, experiment_id, metadata_dict):
        engine = create_engine(db_url)
        Session = sessionmaker(bind=engine)
        session = Session()
        for key, value in metadata_dict.items():
            session.add(ExpMetadata(experiment_id=experiment_id, key=key, value=str(value)))
        session.commit()
        session.close()
        engine.dispose()

    def get_unique_experiment_name(self, base_name, session):
        """
        This function checks if an experiment name exists in the database and, if it does,
        increments a number appended to the name by 1.
        """
        # Regular expression to find a number at the end of the name
        regex = re.compile(r'(.*)_(\d+)$')
        
        # Query the database for names like the base name
        similar_names = session.query(Experiment.name).filter(
            Experiment.name.ilike(f"{base_name}%")
        ).all()
        
        max_number = -1
        for (name,) in similar_names:
            match = regex.match(name)
            if match:
                # If the name matches the pattern, update max_number
                prefix, number = match.groups()
                if prefix.lower() == base_name.lower():
                    max_number = max(max_number, int(number))
            elif name.lower() == base_name.lower():
                # If the name is exactly base_name, ensure we at least get _1
                max_number = max(max_number, 0)
        
        # If max_number is 0, the base name is unique and can be used as is
        # Otherwise, append _{max_number+1} to make it unique
        return f'{base_name}_0' if max_number < 0 else f"{base_name}_{max_number + 1}"
    
    def create_experiment(self, db_url, backup_db_url, name, start_time=None):
        self.db_url = db_url
        self.backup_db_url = backup_db_url
        try:
            experiment_id, name = self._create_experiment_direct(db_url, name, start_time)
        except OperationalError:
            logging.warning("Failed to connect to the main database. Using the backup database.")
            self.db_url = backup_db_url
            experiment_id, name = self._create_experiment_direct(backup_db_url, name, start_time)
        self.main_experiment_id = experiment_id
        if self.db_url != self.backup_db_url:
            backup_experiment_id, _ = self._create_experiment_direct(backup_db_url, name, start_time)
            self.backup_experiment_id = backup_experiment_id
        else:
            backup_experiment_id = experiment_id
            DBWorker.backup_experiment_id = experiment_id
        self.experiment_id = experiment_id
        return experiment_id, backup_experiment_id, name

    def _create_experiment_direct(self, db_url, name, start_time=None):
        if 'sqlite' in db_url:
            engine = create_engine(db_url)
        else:
            engine = create_engine(db_url, connect_args={'connect_timeout': 10})
        Session = sessionmaker(bind=engine)
        session = Session()
        Base.metadata.create_all(engine)
        successfully_created = False
        experiment = session.query(Experiment).filter_by(name=name).first()
        if experiment is not None:
            experiment_id = experiment.experiment_id
            successfully_created = True
        while not successfully_created:
            try:
                if start_time is None:
                    start_time = datetime.now()
                name = self.get_unique_experiment_name(name, session)
                new_experiment = Experiment(name=name, start_time=start_time)
                session.add(new_experiment)
                session.commit()
                experiment_id = new_experiment.experiment_id
                successfully_created = True
            except (psycopg2.errors.UniqueViolation, IntegrityError):
                session.rollback()
        session.close()
        engine.dispose()
        return experiment_id, name
    
    def experiment_end(self):
        self._experiment_end(self.db_url, self.experiment_id)
        self._experiment_end(self.backup_db_url, self.backup_experiment_id)

    def _experiment_end(self, db_url, experiment_id):
        engine = create_engine(db_url)
        Session = sessionmaker(bind=engine)
        session = Session()
        experiment = session.query(Experiment).get(experiment_id)
        if experiment is not None:
            experiment.end_time = datetime.now()
            session.commit()
        session.close()
        engine.dispose()

    def get_component_id(self, component_name):
        try:
            main_component_id = self._get_component_id(self.db_url, component_name)
        except OperationalError:
            main_component_id = None
        _ = self._get_component_id(self.backup_db_url, component_name)
        return main_component_id

    def _get_component_id(self, db_url, component_name):
        engine = create_engine(db_url)
        Session = sessionmaker(bind=engine)
        session = Session()
        component = session.query(Component).filter_by(name=component_name).first()           
        # If not, create it
        if component is None: 
            new_component = Component(name=component_name)
            session.add(new_component)
            session.commit()
            component_id = new_component.component_id  
            
        else:
            component_id = component.component_id
        session.close()
        engine.dispose()
        return component_id



from .register import RegisterMeta 
from .interfacegenerator import InterfaceGeneratorMeta


class HyperParser(RegisterMeta):
    """
    Creates a register for hyper-parametrised classes that could be 
    optimised. This is the metaclass for the AbstractParametered abstract
    base class. 
    Additionally to the register it attaches the hyper-parameters to cls
    """
    # _register = {}

    def __new__(mcl, name, bases, class_dict):
        cls = super().__new__(mcl, name, bases, class_dict)
        cls.harvest_parameters()
        return cls


class AbstractHyperParser(InterfaceGeneratorMeta, HyperParser):
    """
    The AbstractHyperParser allows the class using this metaclass to have
    both HyperParser and ABC functionality.
    This is so abstractclassmethod and abstractmethod decorators can be used
    on the classes.
    """
    pass


from abc import ABC, abstractmethod

from .experimentrunner import ExperimentRunner

class ScriptGenerator(ABC):
    """Generates the script to be run for HPOptimisation
    """
    @abstractmethod
    def generate_script(self, hp_cmd_str: str) -> str:
        """The function that generates the script. Will either
        return a file path or the string to be executed locally.

        :param hp_cmd_str: the string with arguments from HPGenerator
        :type hp_cmd_str: str
        :return: a file path or local execution string
        :rtype: str
        """
        pass


class ScriptRunner(ExperimentRunner):
    def __init__(self, script_gen: ScriptGenerator):
        self._generator = script_gen


from typing import Any

import time
from functools import wraps
import numpy as np

from . import SingleEntity, HybridSingleEntity
from ..core_imports import Lock

class Observer:
    """An Observer from the observer pattern. Uses a name as an id.

    Define an updating interface for objects that should be notified of
    changes in a subject.

    To save space only one observer for each subject is created (quasi singleton)
    When observers are created if one already exists with the name space then 
    the existing observer is given 
    """
    _pending = set()
    _entities = {}

    def __init__(self, name: str, owner: object):
        """Constructor of Observer
        
        :param name: This Observer's name
        :type name: str
        """
        assert isinstance(name, str), f'name: {name} must be a string'
        self._state = None
        self._name = name
        self._owner = owner
        self._lock = Lock()
        self._num_access = 1 if not hasattr(self, '_num_access') else self._num_access
        self._accessed = self._num_access

    def __new__(cls, name:str, owner: object, *args, **kwargs):
        """
        On creation attempts to attach Observer to Subject

        If Subject does not already exist adds Observer to pending
        """
        self = cls.create_new(name, owner, *args, **kwargs)
        try:
            Subject._entities[name].attach(self)
        except KeyError:
            cls._pending.add(self)
        return self
    
    @classmethod
    def create_new(cls, name:str, owner: object, *args, **kwargs):
        """
        Creates a new Observer and attaches it to the Subject
        """
        key = (name, owner)
        if key in cls._entities:
            #print(f'Observer {name} got')
            #entity has been created
            self = cls._entities[key]
            #Increment number of times to access observer
            self._num_access += 1
            return self
        else:
            #print(f'Observer {name} created')
            #create the entity
            new = super().__new__(cls)
            #add entity to the entities dictionary
            cls._entities[key] = new
            return new

    @property
    def name(self) -> str:
        """Gets the Observers name (read only)
        
        :return: The name of this Observer
        :rtype: str
        """
        return self._name

    @property
    def state(self) -> Any:
        """Get Observer's state
        
        :return: Return the Observer state
        :rtype: Any
        """
        counter = 0
        while self.accessed:
            time.sleep(0.01)
            counter += 1
            if counter > 40000:
                raise ValueError(f'Observer {self} has been waiting for {counter*0.01} seconds')
            if self._owner._error_event.is_set():
                raise ValueError(f'Error event erroring out of Observer: {self}')
        self._accessed += 1
        ret_val = self._state
        return ret_val

    def _set_state(self, value: Any):
        """Set the Observer's state
        Should only be called by Subject       
        :param value: The new state for the Observer
        :type value: Any
        """
        # self._lock.acquire()
        self._accessed = 0
        self._state = value
        # self._lock.release()

    def __repr__(self) -> str:
        """Representation of Observer
        
        :return: string representation of Observer
        :rtype: str
        """
        return f"Observer('{self.name}', {self._owner})"
    @property
    def accessed(self):
        return self._accessed == self._num_access

class Subject(SingleEntity):
    """A ConcreteSubject from the Observer pattern. Uses a name as an id.
    """
    _entities = {}

    def __init__(self, name: str, owner: object):
        """Constructor of Subject
        
        :param name: Name of subject
        :type name: str
        """
        assert isinstance(name, str), f'name: {name} must be a string'
        self._observers = set()
        self._owner = owner
        self._state = None
        self._name = name
        self._lock = Lock()
        self.attach_pending_observers()
    
    def attach_pending_observers(self):
        temp = Observer._pending.copy()
        for observer in Observer._pending:
            if self.name == observer.name:
                self.attach(observer)
                temp.remove(observer)
        if temp != Observer._pending:
            Observer._pending = temp

    def attach(self, observer: Observer):
        """Attach an observer to the subject
                
        :param observer: the observer to be attached
        :type observer: Observer
        """
        observer._subject = self
        self._observers.add(observer)

    def detach(self, observer: Observer):
        """Detach an observer from the subject
        
        :param observer: the observer to be detached
        :type observer: Observer
        """
        self._observers.remove(observer)

    def _notify(self):
        """Notify the observers of a change in state
        """
        [observer._set_state(self.state) for observer in self._observers]

    @property
    def name(self) -> str:
        """Getter of the name of the Subject
        
        :return: Name of this agent
        :rtype: str
        """
        return self._name

    #Getter and setter (Concrete Subject?)
    @property
    def state(self) -> Any:
        """Getter of subject state
        
        :return: The subject state
        :rtype: Any
        """
        return self._state

    @state.setter
    def state(self, arg: Any):
        """Setter of the subject state, also notifies all observers
        
        :param arg: The subjects new state
        :type arg: Any
        """
        
        while self.check_locks():
            time.sleep(0.01)
        if any([not x._owner.is_alive() for x in self._observers]):
                self._state = arg
                return
        self._assign_state_notify(arg)    

    def _assign_state_notify(self, arg: Any):
        """Assigns the state of the subject and notifies all observers
        
        :param arg: The new state of the subject
        :type arg: Any
        """
        #print(f'Subject {self._name}: {arg}')
        self._state = arg
        self._notify()

    def initialise_state(self, arg: Any, increment: bool = False):
        """Initialise the state of the subject
        
        :param arg: The initial state of the subject
        :type arg: Any
        """
        # self._lock.acquire()
        self._assign_state_notify(arg)
        if increment:
            for observer in self._observers:
                observer._accessed += 1
        # self._lock.release()

    def check_locks(self)->bool:
        #note we could just detach observers as a matter of course when we end a thread lol.
        return any([(not x.accessed) and x._owner.is_alive() for x in self._observers])


    def release_all_locks(self):
        [x._lock.release() for x in self._observers if x._lock.locked()]

    def __repr__(self) -> str:
        """String representation of the Subject
        
        :return: The string representaion of this subject.
        :rtype: str
        """
        return f"Subject('{self.name}', {self._owner})"

    @classmethod
    def verify_obs_subs(cls):
        """
        All pending observers have been removed and every subject has at least one observer
        """
        if len(Observer._pending) > 0:
            raise ValueError(f"{Observer._pending} is not empty")

        for name in cls._entities:
            #Check no subject has no observer
            if not cls._entities[name]._observers:
                raise ValueError(
                    f'{cls._entities[name]} has no attached observer')


import pexpect
import logging
import sys
import re
import time
import pathlib
import pysftp
from stat import S_ISDIR, S_ISREG
from typing import List, Union
from threading import Event, Lock

ERRORLIMIT = 60

def make_dir(dir_path: pathlib.Path):
    """Make dir if it doens't exist and don't error if it does

    :param dir_path: path to create
    :type dir_path: pathlib.Path
    """    
    try:
        dir_path.mkdir()
    except FileExistsError:
        pass


def check_file_times(from_file:str, to_file:str) -> bool:
    """Check if from_file is newer than to_file

    :param from_file: the source file
    :type from_file: str
    :param to_file: the destination file
    :type to_file: str
    :return: True if from_file is newer than to_file, False otherwise
    :rtype: bool
    """    
    from_time = from_file.lstat().st_mtime if isinstance(from_file,pathlib.Path)\
                else from_file.st_mtime
    to_time = to_file.lstat().st_mtime if isinstance(to_file,pathlib.Path)\
                else to_file.st_mtime
    return int(from_time) > int(to_time)

def check_create(file_path: pathlib.Path) -> bool:
    """Check if file_path exists and create it if it doesn't

    :param file_path: the path to check
    :type file_path: pathlib.Path
    :return: True if file_path exists, False otherwise
    :rtype: bool
    """    
    if not file_path.exists():
        paths_to_create = [file_path]
        while not paths_to_create[0].parent.exists():
            paths_to_create.insert(0, paths_to_create[0].parent)
        [f.mkdir() for f in paths_to_create if not f.suffix]
        return False
    return True

class SFTP(pysftp.Connection):
          
    def __init__(self, host=None, username=None, private_key=None,
                 password=None, port=22, private_key_pass=None, ciphers=None,
                 log=False, cnopts=None, default_path=None):
        if cnopts is None:
            cnopts, hostkeys = self.hostkey_resolution(host)
        super().__init__(host, username, private_key, password, port,
                         private_key_pass, ciphers, log, cnopts, default_path)
        if hostkeys != None:
            print("Connected to new host, caching its hostkey")
            hostkeys.add(host, self.remote_server_key.get_name(),
                         self.remote_server_key)
            hostkeys.save(pysftp.helpers.known_hosts())

    def hostkey_resolution(self, host):
        cnopts = pysftp.CnOpts()
        hostkeys = None
        if cnopts.hostkeys.lookup(host) == None:
            print("New host - will accept any host key")
            # Backup loaded .ssh/known_hosts file
            hostkeys = cnopts.hostkeys
            # And do not verify host key of the new host
            cnopts.hostkeys = None
        return cnopts, hostkeys
    
    def put_r(self, local, remote='.', preserve_mtime=True, confirm=True,
              include=set()):
        local = local if isinstance(local,
                                    pathlib.Path) else pathlib.Path(local)
        if not self.exists(remote):
            self.makedirs(remote)
        if local.is_dir():
            for f in local.iterdir():
                if (set(f.parts) & include):
                    remote_path = remote + '/' + f.name
                    if f.is_dir():
                        if not self.exists(remote_path):
                            self.makedirs(remote_path)
                        self.put_r(f, remote_path, preserve_mtime, confirm,
                                   include)
                    else:
                        if not (self.exists(remote_path)) or check_file_times(
                                f, self.stat(remote_path)):
                            print(f'Copying {f} to {remote_path}')
                            self.put(f, remote_path,
                                     preserve_mtime=preserve_mtime,
                                     confirm=confirm)
        else:
            if not (set(local.parts) & include):
                return
            remote_path = remote + '/' + local.name
            if not (self.exists(remote_path)) or check_file_times(
                    local, self.stat(remote_path)):
                print(f'Copying {local} to {remote_path}')
                self.put(local, remote_path, preserve_mtime=preserve_mtime,
                         confirm=confirm)

    def get_r(self, remote, local, preserve_mtime=True, include=set()):
        local = local if isinstance(local,
                                    pathlib.Path) else pathlib.Path(local)
        make_dir(local)
        for entry in self.listdir_attr(remote):
            #remote must use posix file system
            remote_path = remote + '/' + entry.filename
            if not (set(remote_path.split('/')) & include):
                continue
            #local can be either windows or posix
            local_path = local / entry.filename
            mode = entry.st_mode
            if S_ISDIR(mode):
                make_dir(local_path)
                self.get_r(remote_path, local_path, preserve_mtime, include)
            elif S_ISREG(mode):
                if not (local_path.exists()) or check_file_times(
                        entry, local_path):
                    print(f'Copying {remote_path} to {local_path}')
                    self.get(remote_path, local_path,
                             preserve_mtime=preserve_mtime)
                    
class FilePathGenerator:
    def __init__(self, local_path: pathlib.Path, remote_path: pathlib.Path):
        if not isinstance(local_path, pathlib.Path):
            local_path = pathlib.Path(local_path)
        self._local_path = local_path
        if not isinstance(remote_path, pathlib.Path):
            remote_path = pathlib.Path(remote_path)
        self._remote_path = remote_path

    def generate_local_remote_path(self, file_name:str):
        stem = pathlib.Path(*self._local_path.parts[-2:])
        local_file = self._local_path / file_name
        check_create(local_file.parent)
        remote_path = self._remote_path / stem / file_name
        return local_file, remote_path

    def __call__(self, file_num: int) -> List[pathlib.Path]:
        return self.generate_local_remote_path(file_num)

class PexpectTimeoutError(Exception):
    pass

class PexpectConnectionError(Exception):
    pass

class PexpectSFTPClient:
    """An SFTP client that uses pexpect to execute commands on the remote host.

    This was created so that ssh configurations can be used to connect to remote hosts. 
    The issue with using paramiko is that it is poorly suited to complex ssh configurations.
    It also allows multiple commands in a single session which can be very useful for setting up
    environment variables before executing the sbatch command.
    """
    _stop = Event()
    _counter = 0
    _counter_lock = Lock()
    def __init__(self, config_shortname:str, ssh_prompt:str):
        """Initialise the PexpectSFTPClient.

        :param config_shortname: The host name of the ssh configuration
        :type config_shortname: str
        :param ssh_prompt: The default ssh prompt on your remote host. Caution: it is different on different hosts.
        :type ssh_prompt: str
        """        
        self.config_shortname = config_shortname
        self.ssh_cmd = f"ssh {config_shortname}"
        self.sftp_cmd = f"sftp {config_shortname}"
        self.sftp_prompt = r'sftp> '
        self.ssh_prompt = ssh_prompt
        
        # # Set up logging
        logging.basicConfig(level=logging.INFO)
        self.logger = logging.getLogger(__name__)
        self.logger.disabled = True

    def check_lock(self, e = None, child = None):
        with PexpectSFTPClient._counter_lock:
            if e is None:
                #Reset counter
                PexpectSFTPClient._counter = 0
            else:
                ssh_output = child.before
                #Multiplex not happening
                error_message = ''
                if 'passcode' in ssh_output: # or 'restricted to employees, students, or individuals authorized' in ssh_output: issues with rangpur connectivity mean this is not so good.
                    PexpectSFTPClient._counter = ERRORLIMIT
                    error_message = 'MFA - Passcode required'
                #Wrong host
                elif 'Could not resolve hostname' in ssh_output:
                    PexpectSFTPClient._counter = ERRORLIMIT
                    error_message = 'Could not resolve hostname'
                else:
                    PexpectSFTPClient._counter += 1
                if PexpectSFTPClient._counter >= ERRORLIMIT:
                    error_message = f'Error limit reached' if not error_message else error_message
                    PexpectSFTPClient._stop.set()
                    raise PexpectConnectionError(f'{error_message} : {e}')
                
                
    def _execute_ssh_command(self, command: str)->str:
        """Execute a command on the ssh machine

        :param command: The command you wish to execute e.g ls -l
        :type command: str
        :raises PexpectTimeoutError: The command has timed out 
        :raises PexpectConnectionError: There was an error connecting to the remote host
        :return: The stdout of the command
        :rtype: str
        """
        ssh_agent_pid = None
        if self._stop.is_set():
            raise PexpectConnectionError('SSH connection failed in another thread') 
        try:
            result = None
            child = pexpect.spawn(self.ssh_cmd, encoding='utf-8', timeout=60)
            # Uncomment the below line if you need to debug your connection
            # child.logfile = sys.stdout  # Logs the output to stdout for debugging
        
            # This may or may not work on your remote machine.
            # Get the PID of the ssh-agent to kill it before exiting.
            # try:
            #     child.expect("Agent pid (\d+)", timeout=20)
            #     output = child.after  # This contains the matched output
            #     # Extract the PID using regular expression
            #     match = re.search(r'Agent pid (\d+)', output)
            #     if match:
            #         ssh_agent_pid = int(match.group(1))
            #     else:
            #         print('No agent PID found')
            # except pexpect.TIMEOUT:
            #     print('No agent PID found')
            #     self.check_lock()            
            # Execute the command
            try:
                child.expect(self.ssh_prompt)
                # output = child.before.strip()
                #This is specific to me and I should make this a function in an extended class rather than here
                if command.split(' ')[0] == 'sbatch':
                    child.sendline('source ~/.bash_profile')
                    child.expect(self.ssh_prompt)
                # Run command specified
                child.sendline(command)
                child.expect(self.ssh_prompt)
                # Get the output excluding the first and last line
                result = child.before.strip().split('\n')
                result = '\n'.join(result[1:-1])
                child.sendline(f'pkill -u uqlsalt -f ssh-agent')
                child.expect(self.ssh_prompt)
                time.sleep(1)
                child.sendline('exit\r')
                child.expect(pexpect.EOF)
                child.close()
                self.check_lock()
                return result
            # Handle timeout. Usually failed ssh connection
            except (pexpect.EOF, pexpect.TIMEOUT) as e:
                self.logger.error(f"Error occurred: {e}")
                self.check_lock(e, child)
                child.close()
                return None
        # I'm not sure if this actually catches anything as I haven't hit it but it's there just in case
        except (pexpect.EOF, pexpect.TIMEOUT) as e:
            self.logger.error(f"Error occurred: {e}")
            self.check_lock(e, child)
        except pexpect.exceptions.ExceptionPexpect as e:
            self.logger.error(f"Error occurred: {e}")
            return None
        
        
    def _execute_sftp_command(self, command: str) -> str:
        """Executes SFTP for file exchange

        :param command: The sftp command to be executed
        :type command: str
        :return: The output of the sftp command
        :rtype: str
        """
        if self._stop.is_set():
            raise PexpectConnectionError('SSH connection failed in another thread')       
        child = pexpect.spawn(self.sftp_cmd, encoding='utf-8', timeout=60)
        # child.logfile = sys.stdout  # Logs the output to stdout for debugging
        
        try:
            child.expect(self.sftp_prompt)
            child.sendline(command)
            child.expect(self.sftp_prompt)
            result = child.before.strip()
            child.sendline('bye')
            child.expect(pexpect.EOF)
            self.check_lock()
            child.close()
            return result
        except (pexpect.EOF, pexpect.TIMEOUT) as e:
            self.logger.error(f"Error occurred: {e}")
            self.check_lock(e, child)
            child.close()
            return None
            
        
    def exists(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a file or path exists

        :param path: The dir or file you wish to check
        :type path: Union[str, pathlib.Path]
        :return: true if path exists, false otherwise
        :rtype: bool
        """                
        result = self._execute_ssh_command(f'test -e {path} && echo EXISTS || echo NO')
        #Recheck if None (feels dangerous.)
        if result is None:
            return False
        return "EXISTS" in result

    def makedirs(self, remote_directory: str):
        """Make all directories in remote_directory on remote host

        :param remote_directory: The path you wish to create
        :type remote_directory: str
        """        
        directories = remote_directory.strip().split('/')
        for i in range(1, len(directories) + 1):
            path = '/'.join(directories[:i])
            if not self.exists(path):
                self.logger.info(f"Creating directory: {path}")
                self._execute_sftp_command(f'mkdir {path}')

    def put(self, local_path: Union[str, pathlib.Path], remote_path: Union[str, pathlib.Path]):
        """Put file from local host onto remote host

        :param local_path: path to file on local
        :type local_path: Union[str, pathlib.Path]
        :param remote_path: path to file on remote
        :type remote_path: Union[str, pathlib.Path]
        """        
        self.logger.info(f"Uploading {local_path} to {remote_path}")
        self._execute_sftp_command(f'put {local_path} {remote_path}')

    def execute(self, command: str) -> str:
        """Execute a command on the remote host

        :param command: The command to execute on the remote host
        :type command: str
        :return: The output of the command on the remote host
        :rtype: str
        """        
        self.logger.info(f"Executing command: {command}")
        return self._execute_ssh_command(command)



class RegisterMeta(type):
    """
    The register metaclass adds a dictionary to class that uses the metaclass. 
    Automates registering classes at class definition, which is typically done
    manually. (e.g gym when creating custom environments)

    Metaclasses that inherit from register need to redefine the _register 
    variable if separate registers should be maintained.

    This is typically useful for factory classes/functions so that the desired
    class can be specified with a string
    """

    def __new__(mcl, name, bases, class_dict):
        """
        Adds any new classes to _register when the class is defined and
        uses the register metaclass or any of it's children
        """
        #Hopefully this will remove the need to add a register to each new metaclass.
        if (not hasattr(mcl,"_register")):
            mcl._register = {}
        cls = type.__new__(mcl, name, bases, class_dict)
        if (not cls.__dict__.get("__abstractmethods__")):
            if cls.__name__ in mcl._register:
                raise ValueError(f"Class with this name: {cls.__name__} already exists in this registry")
            else:
                mcl._register[cls.__name__] = cls
        return cls


class SingletonMeta(type):
    """
    Standard singleton pattern
    """
    _instance = None
    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
            return cls._instance
        return cls._instance


from abc import ABC, abstractmethod

from .experiment import AbstractExperiment

class AbstractComponentFactory(ABC):
    """Constructs components
    
    Register for each factory records which classes are constructed by
    the factory interface. The 
    """
    _register = []

    def __init__(self, experiment: AbstractExperiment):
        self._components = experiment.components
        self._exp_args = experiment._exp_args
        self._experiment = experiment
        self.build()

    @abstractmethod
    def build(self):
        pass



from typing import Any


class TypedDict(dict):
    """
    A dictionary that only allows values of value_class, 
    keys of type str, and no overriding of values. All keys are cast to 
    lowercase.
    """
    def __init__(self, value_klass: Any):
        """Constructor of TypedDict
        
        :param value_klass: [description]
        :type value_klass: Any
        """
        self._value_klass = value_klass

    def __getitem__(self, key: str) -> Any:
        """Gets the value with key
        
        :param key: The key of the value
        :type key: str
        :return: The value at key
        :rtype: Any (value_klass)
        """
        return super().__getitem__(key.lower())

    def __setitem__(self, key: str, value: Any):
        """Set the key value pair. Value must be of 
        
        :param key: The key of the value
        :type key: str
        :param value: The value for the key
        :type value: Any (value_klass)
        """
        if not isinstance(key, str):
            raise TypeError("Key must be of type str for TypedDict")
        if not isinstance(value, self._value_klass):
            raise TypeError(
                f"value must be an instance of {self._value_klass} for TypedDict"
            )
        return super().__setitem__(key.lower(), value)

    def __delitem__(self, key: str) -> bool:
        """Deletes the entry at key
        
        :param key: The key of the subject to be deleted
        :type key: str
        :return: whether or not item is deleted
        :rtype: bool
        """
        return super().__delitem__(key.lower())

    def __contains__(self, item: str) -> bool:
        """Checks if the dictionary contains the item

        :param item: key of the item we're looking for
        :type item: str
        :return: whether or not item is in the dictionary keys
        :rtype: bool
        """
        if not isinstance(item, str):
            raise TypeError(
                f"{self.__class__}'s keys can only contain type({str})")
        return item.lower() in self.keys()


import optuna
import numpy as np
from typing import Union

from .hyperparameters import DefaultExperimentHyperParameters
from .experimentrunner import ExperimentRunner
from ..interfaces import AbstractExperiment, HyperParameter
from ..logger import create_experiment_direct
from ..optimiser.sftp import PexpectConnectionError, PexpectTimeoutError

class OptunaObjective:
    def __init__(self, experiment: AbstractExperiment, exp_runner: ExperimentRunner, opt_kwargs:dict, storage:str=None):
        """An objective function class for Optuna to optimise. This class
        is used to generate the hyperparameters for the experiment. It does this
        by parsing the experiments for hyperparameters defined by set_up_hyperparameters.

        :param experiment: The experiment to be optimised
        :type experiment: AbstractExperiment
        :param exp_runner: a class that runs experiment, must have run method that returns a float, defaults to None
        :type exp_runner: ExperimentRunner, optional
        """
        opt_kwargs = self.convert_opt_kwargs(opt_kwargs)
        self._opt_args = {'--strategy': experiment.__name__, **opt_kwargs}
        self._name = self._opt_args.pop('--name', None)
        if self._name is None:
            self._name = experiment.__name__  
        self._experiment = experiment
        self._exp_runner = exp_runner
        self._hp_dict = {}
        self._exp_hps = DefaultExperimentHyperParameters(experiment)
        self._storage = storage

    def objective(self, trial: optuna.trial.Trial) -> float:
        """The objective function for Optuna to optimise

        :param trial: the trial object from Optuna
        :type trial: optuna.trial.Trial
        :return: the result of the experiment
        :rtype: float
        """ 
        self._hp_dict['--name'] = create_experiment_direct(self._name, study_id=trial.study._study_id, db_url=self._storage)[1]
        trial.set_user_attr("name", self._hp_dict['--name'])
        hp_cmd_str = self.parse_hyper_parameters(trial)
        max_retries = 3
        for i in range(max_retries):
            try:
                return self._exp_runner.run(hp_cmd_str,
                                            trial, self._storage)
            except (PexpectConnectionError, PexpectTimeoutError) as e:
                raise e
            except Exception as e:
                print(f'Error occurred: {e}')
                print(f'Retrying {i + 1}/{max_retries}')
                trial.set_user_attr("failed", True)
                raise e
        return None

    def convert_opt_kwargs(self, opt_kwargs: dict) -> str:
        """Convert the optuna kwargs into a string that can be parsed by the experiment

        :param opt_kwargs: the optuna kwargs
        :type opt_kwargs: dict
        :return: a string that can be parsed by the experiment
        :rtype: str
        """
        temp_dict = {}      
        for key, val in opt_kwargs.items():
            if not key.startswith('--'):
                key = f'--{key}'.replace('_', '-')
            temp_dict[key] = val
        return temp_dict

    def parse_hyper_parameters(self, trial:optuna.trial.Trial) -> str:
        """Parse the hyper parameters of the experiment and determine 
        if they should have a suggestion made.

        :param trial: the current optuna trial
        :type trial: optuna.trial.Trial
        :return: the string to be fed to the experiment runner
        :rtype: str
        """        
        remove = []
        for hp_name, hp in self._exp_hps.hyperparameters.items():
            length = self._parse_length(trial, hp_name, hp)
            suggestions = []
            for i in range(length):
                name = hp_name
                if length > 1: name += f'_{i}'
                suggestion = self._make_suggestion(trial, name, hp, i)
                if suggestion not in [None, (), {}, []]:
                    suggestions.append(str(suggestion))
                else:
                    remove.append(name)
            if suggestions not in [None, (), {}, []]:
                self._hp_dict[hp_name] = suggestions
        [self._hp_dict.pop(x, None) for x in remove]
        temp_dict = self._hp_dict.copy()
        [self._hp_dict.pop(x, None) for x in temp_dict if self._hp_dict[x] in [None, (), {}, []]]
        self._hp_dict = {**self._hp_dict, **self._opt_args}
        return self.convert_hp_dict_str()

    def convert_hp_dict_str(self) -> str:
        """Convert the hyperparameter dictionary into a string that can be used by the exprunner CLI

        :return: the experiment parameters as a string for exprunner
        :rtype: str
        """        
        hp_str = ''
        for key, val in self._hp_dict.items():
            if hasattr(val, '__len__') and not isinstance(val, str):
                val = [str(x) for x in val]
                hp_str += f"{key} {' '.join(val)} "
            else:
                hp_str += f"{key} {val} "
        return hp_str

    def _parse_length(self, trial: optuna.trial.Trial, hp_name: str,
                      hyperparameter: HyperParameter) -> int:
        """Determine if parsing should be done for the hyperparameter

        :param trial: the current optuna trial
        :type trial: optuna.trial.Trial
        :param hp_name: the name of the hyperparameter
        :type hp_name: str
        :param hyperparameter: The hyperparameter object
        :type hyperparameter: HyperParameter
        :raises NotImplementedError: errors if hyperparameter type is not accounted for
        :return: returns the length of the hyperparameter
        :rtype: int
        """        
        if type(hyperparameter.length) in (tuple, list):
            if hyperparameter.type == np.ndarray:
                raise NotImplementedError(
                    'Not actually sure what this type is' +
                    'for but you should be now so fix this')
            else:
                assert len(hyperparameter.length) == 2
                return trial.suggest_int(hp_name.strip('--')+"-length",
                                         hyperparameter.length[0],
                                         hyperparameter.length[1])
        else:
            return hyperparameter.length

    def _make_suggestion(self, trial: optuna.trial.Trial, hp_name: str,
                         hyperparameter: HyperParameter, index: int)->Union[int, float, None]:
        """Make a suggestion for the hyperparameter

        :param trial: the current trial
        :type trial: optuna.trial.Trial
        :param hp_name: the name of the hyperparameter
        :type hp_name: str
        :param hyperparameter: the hyperparameter object
        :type hyperparameter: HyperParameter
        :param index: _description_
        :type index: int
        :raises NotImplementedError: errors if type not considered (not numeric)
        :return: The suggestion
        :rtype: Union[int, float, None]
        """        
        hp_name = hp_name.strip('--')
        if hyperparameter.bounds is None:
            return None
        elif hyperparameter.type == int or type(
                hyperparameter.bounds[0]) == int:
            return trial.suggest_int(hp_name, hyperparameter.bounds[0],
                                     hyperparameter.bounds[1])
        elif hyperparameter.type == float or type(
                hyperparameter.bounds[0]) == float:
            return trial.suggest_float(hp_name, hyperparameter.bounds[0],
                                         hyperparameter.bounds[1])
        else:
            raise NotImplementedError(
                f'{hyperparameter} not yet accounted for')


import json
from abc import ABC, abstractmethod
from typing import Callable
import numpy as np
import pathlib
import inspect
import importlib

from ..metaclasses.register import RegisterMeta


class SerialRegister(RegisterMeta):
    """A metaclass for registering classes. Intended for use with serialisable.
    Automates decorating the init method with serialise_init
    """
    _register = {}
    serial_register = {'Serialisable'}
    experiment_register = {'AbstractExperiment'}

    def __new__(mcl, name, bases, class_dict):
        set_bases = {cls.__name__ for cls in bases}
        #intercept initialisers
        if (set_bases & SerialRegister.serial_register):
            SerialRegister.serial_register.add(name)
            if '__init__' in class_dict:
                class_dict['__init__'] = serialise_init(class_dict['__init__'])
        cls = super().__new__(mcl, name, bases, class_dict)
        #Make a set of experiments.
        if (set_bases & SerialRegister.experiment_register):
            cls._arg_parser()
            SerialRegister.experiment_register.add(name)
        return cls


class Serialisable(metaclass=SerialRegister):
    """A serialisable mixin that uses the SerialRegister metaclass
    
    Redefines str and repr and also provides a JSON serialise method.
    """
    def serialise(self, file_path: pathlib.Path = None) -> str:
        """Serialises self
        
        :return: A JSON string that contains all information necessary to reinitialise self
        :rtype: str
        """
        args, kwargs = self.args[:], self.kwargs.copy()
        for i, arg in enumerate(args):
            if hasattr(arg, 'serialise'):
                args[i] = arg.serialise()
        for key, value in kwargs.items():
            if hasattr(value, 'serialise'):
                kwargs[key] = value.serialise()
        j_dict = {
            'class': self.__class__.__name__,
            'args': args,
            'kwargs': kwargs
        }
        if file_path is not None:
            with open(file_path, 'w') as f:
                json.dump(j_dict, f, cls=NumpyEncoder, indent=4)
        return json.dumps(j_dict, cls=NumpyEncoder, indent=4)

    def __str__(self):
        items = [
            str(x) if not isinstance(x, str) else f"'{x}'" for x in self.args
        ] + [
            f'{k}={v}' if not isinstance(v, str) else f"{k}='{v}'"
            for k, v in self.kwargs.items()
        ]
        return f"{self.__class__.__name__}({','.join(items)})"

    __repr__ = __str__


class RegisteredAbstract(type(ABC), SerialRegister):
    """A metaclass for those classes that use both ABC and SerialRegister
    """
    pass


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.int_, np.int16, np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, np.ndarray):
            return {'data': obj.tolist(), 'dtype': str(obj.dtype)}
        elif isinstance(obj, pathlib.Path):
            return {'path': str(obj)}
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif callable(obj): 
            return str(obj) 
        return json.JSONEncoder.default(self, obj)


def inspect_signature(func):
    signature = inspect.signature(func)
    kwonlys = {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }
    return kwonlys


def init_error_checker(args: tuple):
    if hasattr(args[0], 'args') != hasattr(args[0], 'kwargs'):
        raise ValueError(
            f'{args[0]} should either have args AND kwargs or neither')


def combine_kwargs_kwonlys(kwargs: dict, kwonlys: dict):
    if kwonlys:
        for key, val in kwonlys.items():
            if key not in kwargs:
                kwargs[key] = val


def create_args_kwargs_attr(args: tuple, kwargs: dict, kwonlys: dict):
    init_error_checker(args)
    combine_kwargs_kwonlys(kwargs, kwonlys)
    # object.__setattr__ is necessary for classes that have
    # __setattr__ overrides.
    object.__setattr__(args[0], 'args', list(args[1:]))
    object.__setattr__(args[0], 'kwargs', kwargs)
    return args, kwargs


def serialise_init(init):
    """The wrapper of the init function for any subclass of Serialisable

    :param init: The init function of a class that inherits from Serialisable
    :type init: Callable
    """
    def wrapper_serialise_init(*args, **kwargs):
        """Grabs the input arguments of the initialiser
        and makes them attributes of that instance"""
        kwonlys = inspect_signature(init)
        create_args_kwargs_attr(args, kwargs, kwonlys)
        init(*args, **kwargs)

    return wrapper_serialise_init


def check_if_serial(serialised):
    return 'class' in serialised and 'args' in serialised and 'kwargs' in serialised


def numpy_hook(dct):
    if 'dtype' in dct:
        return np.array(dct['data'], dtype=dct['dtype'])
    elif 'path' in dct:
        return pathlib.Path(dct['path'])
    return dct


def deserialise(serialised: str):
    """A function for deserialising Serialisable classes
    
    :param serialised: The output of Serialisable.serialise.
    :type serialised: str
    """
    if not check_if_serial(serialised):
        raise ValueError('Invalid input to deserialise')
    params = json.loads(serialised, object_hook=numpy_hook)
    cls = SerialRegister._register[params['class']]
    args, kwargs = params['args'], params['kwargs']
    for i, arg in enumerate(args):
        if type(arg) == str and check_if_serial(arg):
            args[i] = deserialise(arg)
    for key, value in kwargs.items():
        if type(value) == str and check_if_serial(value):
            kwargs[key] = deserialise(value)
    return cls(*args, **kwargs)


import argparse
from ..interfaces import AbstractExperiment, HyperParameter


class ExperimentHyperParameters:
    _hyperparameters = None

    @property
    def hyperparameters(self):
        return self._hyperparameters


class DefaultExperimentHyperParameters(ExperimentHyperParameters):
    def __init__(self, experiment: AbstractExperiment):
        self._experiment = experiment
        self.get_cmd_args()
        self.get_hyperparameters()

    def get_cmd_args(self):
        self._parser = argparse.ArgumentParser(
            parents=[self._experiment._parser])
        self._cmd_args = list(self._parser._option_string_actions.keys())

    def get_hyperparameters(self):
        self._hyperparameters = {}
        for builder in self._experiment._builders:
            for component in builder._register:
                for k, v in component.hyperparameters.items():
                    hp_name = f'--{component.__name__}-{k}'.replace('_', '-')
                    self._hyperparameters[hp_name] = v


class CustomExperimentHyperParameters(ExperimentHyperParameters):
    def __init__(self, custom_hp_dict: dict):
        """For custom HP optimisation

        Contains cmd-arg as key and tuple containing optimisation bounds and dist as value

        :param custom_hp_dict:  {--key-name - str:
                                (lower bound - Union(int,float),
                                 upper bound - Union(int,float))}
        :type custom_hp_dict: dict
        """
        self.get_hyperparameters(custom_hp_dict)

    def get_hyperparameters(self, custom_hp_dict: dict):
        self._hyperparameters = {}
        for k, v in custom_hp_dict.items():
            self.hyperparameters[k] = HyperParameter(None, bounds=v,
                                                     htype=type(v[0]))



import pathlib
import sys
import os
import argparse
import shutil

from typing import Dict, List

from .. import AbstractExperiment, DatabaseLogger

os.environ['CUDA_LAUNCH_BLOCKING'] = "1"

def check_exps(strat: str) -> bool:
    """Check if an experiment has been defined

    :param strat: the desired experiment strategy
    :type strat: str
    :return: if the experiment has been defined
    :rtype: bool
    """    
    exps = [x for x in AbstractExperiment._register if 'Abstract' not in x]
    return strat in exps


def runexp(experiment: AbstractExperiment, args: Dict[str,str], unknowns:List[str]):
    """Run the experiment with the given arguments

    :param experiment: The experiment to run
    :type experiment: AbstractExperiment
    :param args: The known arguments as defined in if name main block
    :type args: Dict[str,str]
    :param unknowns: The experiment arguments
    :type unknowns: List[str]
    """    
    
    parser = argparse.ArgumentParser(parents=[experiment._parser])
    args = {**args, **vars(parser.parse_args(unknowns))}
    name = args.pop('name', None)
    name = name if name else experiment.__name__
    file_path = pathlib.Path(__file__).parent.parent.parent.parent / 'SB3Experiments' 
    args['file_path'] = file_path if args['path'] == '' else args['path']
    logger = DatabaseLogger(name, exp_metadata_dict=args)
    should_load = False
    if args['path'] == '':
        args.pop('path')
        file_path = pathlib.Path(__file__).parent.parent.parent.parent / 'SB3Experiments' / logger.experiment_name
        try:
            file_path.mkdir(parents=True, exist_ok=False)
        except FileExistsError:
            shutil.rmtree(file_path)
            file_path.mkdir(parents=True, exist_ok=False)
    else:
        file_path = pathlib.Path(args.pop('path')) / "Experiments" / logger.experiment_name
        should_load = file_path.exists()
    experiment = experiment(file_path, args)
    if should_load:
        pass
    else:
        experiment.run()
    experiment.test()


import inspect
from abc import abstractmethod, ABCMeta

class InterfaceGeneratorMeta(ABCMeta):
    def __new__(mcl, name, bases, class_dict):
        """
        Dynamically create a class interface based on the method names defined in that class

        Incorporating a new method of defining IO so that it becomes extensible and generalisable. Name alteration may be needed.
        
        This is to be used with caution as ABCMeta recommends using update_abstractmethods decorator
        
        This could be done as a class factory if that is preferred. e.g:
        def metafactory(cls):
            assert(hasattr(cls, "method_names"))
            assert(type(cls)==ABCMeta)
            [setattr(cls, x, abstractmethod(lambda self: None)) for x in cls.method_names]
            return update_abstractmethods(cls)
        @metafactory
        class YourClass(ABC):
            method_names = ["method_a", "method_b"]
        """
        #First interface level
        if "abstract_methods" in class_dict:
            cls = super().__new__(mcl, name, bases, {**class_dict, \
                                                    **{method_name:abstractmethod(lambda self, **kwargs:None) \
                                                        for method_name in class_dict["abstract_methods"]}})
        else:
            cls = super().__new__(mcl, name, bases, class_dict)
        return cls



from abc import ABC, abstractmethod

class ExperimentRunner(ABC):
    @abstractmethod
    def run(self, hp_cmd_str: str) -> float:
        """Runs the experiment and returns the result

        :param hp_cmd_str: the string with arguments from HPGenerator
        :type hp_cmd_str: str
        :return: the result of the experiment
        :rtype: float
        """
        pass



#################################################################################################################
# This is a fresh class to formalise the IO a little better so that it is easy to keep track of which           #
# subfunction each IO belongs to as the entire point of this OS is to control the flow of messages so that      #
# we can interrupt implemented algorithms and extract the necessary information to have our own segment work    #
# and alter the original algorithm i.e a truly modular solution.                                                #
# Important features:                                                                                           #
# 1) Automatic graph generation to ensure that we have completed loops                                          #
# 2) extensible to any number of IO segmentation                                                                #
# This class should implement the gating                                                                        #
#################################################################################################################
import inspect
import ast
import re
import random
import string

from typing import Dict, List, Tuple, Union, Set, Callable, Any
from ..patterns import Subject, Observer
from ..utility import listify_inputs

def get_operation(op: ast.operator)->str:
    """Get the operation of an ast.operator

    :param op: the operator
    :type op: ast.operator
    :return: the string representaion of the operator
    :rtype: str
    """
    if isinstance(op, ast.Add):
        return "+"
    elif isinstance(op, ast.Sub):
        return "-"
    elif isinstance(op, ast.Mult):
        return "*"
    elif isinstance(op, ast.Div):
        return "/"
    elif isinstance(op, ast.Pow):
        return "**"
    elif isinstance(op, ast.Mod):
        return "%"
    elif isinstance(op, ast.FloorDiv):
        return "//"
    elif isinstance(op, ast.BitAnd):
        return "&"
    elif isinstance(op, ast.BitOr):
        return "|"
    elif isinstance(op, ast.BitXor):
        return "^"
    elif isinstance(op, ast.LShift):
        return "<<"
    elif isinstance(op, ast.RShift):
        return ">>"
    elif isinstance(op, ast.MatMult):
        return "@"
    elif isinstance(op, ast.Lt):
        return "<"
    elif isinstance(op, ast.Gt):
        return ">"
    elif isinstance(op, ast.LtE):
        return "<="
    elif isinstance(op, ast.GtE):
        return ">="
    elif isinstance(op, ast.Not):
        return "not"

def extract_attribute(value: ast.Attribute)->str:
    """Extract the attribute from an ast.Attribute

    :param value: the value
    :type value: ast.Attribute
    :return: the string representation of the attribute
    :rtype: str
    """    
    values = []
    while isinstance(value, ast.Attribute):
        values.append(value.attr)
        value = value.value
    return ".".join(reversed(values))

def get_innermost_arg(call, args=None):
    if args is None:
        args = []
    if isinstance(call, ast.Call):
        # If the function being called is itself a call, recurse on that
        # Also, add the arguments of the call to the args list
        if isinstance(call.func, ast.Attribute):
            if isinstance(call.func.value, ast.Attribute):
                #print(f'appending {call.func.value.attr} in from call.func.value.attr')
                args.append(call.func.value.attr)
        for arg in call.args:
            if isinstance(arg, ast.Attribute):
                #print(f'appending {arg.attr} in args atteibute')
                args.append(arg.attr)
            elif isinstance(arg, ast.Call):
                get_innermost_arg(arg, args)
            elif isinstance(arg, ast.Starred):
                if isinstance(arg.value, ast.Attribute):
                    #print(f'appending {arg.value.attr} in args starred')
                    args.append(arg.value.attr)
        for keyword in call.keywords:
            if isinstance(keyword.value, ast.Attribute):
                #print(f'appending {keyword.value.attr} in keyword attr')
                args.append(keyword.value.attr)
            elif isinstance(keyword.value, ast.Call):
                get_innermost_arg(keyword.value, args)
        return get_innermost_arg(call.func, args)
    elif isinstance(call, ast.Attribute):
        # If the function being called is an attribute access, recurse on the value the attribute is being accessed on
        return get_innermost_arg(call.value, args)
    else:        
        # Otherwise, return the argument to the innermost call
        return [f'self.{arg}' for arg in args]
    
def get_value(value: ast.AST, cls:object) -> str:
    """Get the value of an ast.AST

    :param value: the value
    :type value: ast.AST
    :return: the string representation of the value
    :rtype: str
    """
    if isinstance(value, ast.Constant):
        return value.value
    elif isinstance(value, ast.Attribute):
        #print(value.attr)
        base = get_value(value.value, cls)
        return f"{base}.{value.attr}"
        
    elif isinstance(value, ast.Name):
        return value.id
    elif isinstance(value, ast.Call):
        # Assume that the function is not a method/function 
        # that has additional IO..., for now
        ret_value = ", ".join(get_innermost_arg(value))
        return ret_value
    elif isinstance(value, ast.Tuple):
        ret_value = ", ".join([get_value(arg,cls) for arg in value.elts])
        return ret_value
    elif isinstance(value, ast.DictComp):
        """The function needs to take in the instance. We then need to get all the values being pulled by the generator.
        We then get all of the values in the generator and generate a string that is all of those values separated by commas or something...
        """
        print("Warning this is parsing a DictComp and the implementation is not robust...")
        # key = value.key
        generators = value.generators
        # value = value.value
        nesting = get_value(generators[0],cls).split('.')[1:]
        for i,nest in enumerate(nesting):
            if i == 0:
                values = getattr(cls, nest)
            else:
                values = getattr(values, nest)
        return ", ".join([f"self.{value}" for value in values])
    
    elif isinstance(value, ast.comprehension):
        """This is a bandaid and needs to be streamlined way more..."""
        print("Warning this is parsing a comprehension and the implementation is not robust...")

        target = value.target
        iter = value.iter
        ifs = value.ifs
        # temp = f"{get_value(target,cls)} in {get_value(iter,cls)}"
        # if ifs:
        #     temp += f" if {get_value(ifs[0],cls)}"
        return f"{get_value(iter,cls)}"
    
    elif isinstance(value, ast.BinOp):
        left = value.left
        right = value.right
        return f"{get_value(left,cls)} {get_operation(value.op)} {get_value(right,cls)}"
    elif isinstance(value, ast.UnaryOp):
        operand = value.operand
        return f"{get_operation(value.op)} {get_value(operand,cls)}"
    elif isinstance(value, ast.Compare):
        left = value.left
        right = value.comparators[0]
        return f"{get_value(left,cls)} {get_operation(value.ops[0])} {get_value(right,cls)}"
    elif isinstance(value, ast.BoolOp):
        left= value.values[0]
        right = value.values[1]
        return f"{get_value(left,cls)} {get_operation(value.op)} {get_value(right,cls)}"
    elif isinstance(value, ast.Dict):
        keys = value.keys
        values = value.values
        return ', '.join([f'{get_value(key,cls)}' for key in keys] + [f'{get_value(value,cls)}' for value in values])
    elif isinstance(value, ast.Subscript):
        container = get_value(value.value, cls)
        subs = value.slice
        return f"{get_value(container,cls)}[{get_value(subs,cls)}]"
    elif not isinstance(value, ast.AST):
        return str(value)
    elif isinstance(value, ast.Slice):
        lower = value.lower
        upper = value.upper
        step = value.step
        if lower:
            lower = get_value(lower,cls)
        if upper:
            upper = get_value(upper,cls)
        if step:
            step = get_value(step,cls)
        return f"{lower}:{upper}:{step}"
    elif isinstance(value, ast.List):
        elts = value.elts
        return f"[{', '.join([str(get_value(elt,cls)) for elt in elts])}]"
    else:
        raise NotImplementedError(f"Value {value} not implemented")
    
# Instead of converting to a string we return a list of ast.AST
def get_value_v2(value: ast.AST)->List[ast.AST]:
    """Get the value of an ast.AST

    :param value: the value
    :type value: ast.AST
    :return: List of ASTs
    :rtype: List[ast.AST]
    """
    if isinstance(value, ast.Constant):
        return [value]
    elif isinstance(value, ast.Attribute):
        return [value]
    elif isinstance(value, ast.Name):
        return [value]
    elif isinstance(value, ast.Call):
        return [value]
    elif isinstance(value, ast.BinOp):
        left = value.left
        right = value.right
        return get_value_v2(left) + get_value_v2(right)
    elif isinstance(value, ast.UnaryOp):
        operand = value.operand
        return get_value_v2(operand)
    elif isinstance(value, ast.Compare):
        left = value.left
        right = value.comparators[0]
        return get_value_v2(left) + get_value_v2(right)
    elif isinstance(value, ast.BoolOp):
        left= value.values[0]
        right = value.values[1]
        return get_value_v2(left) + get_value_v2(right)

# Parses the list output by get_value_v2 and returns a list of strings if the value is a Name or Attribute
def process_values(values: List[ast.AST])->List[str]:
    """Process the values

    :param values: the values
    :type values: List[ast.AST]
    :return: the string representation of the values
    :rtype: List[str]
    """
    processed_values = []
    for value in values:
        if isinstance(value, ast.Name):
            processed_values.append(value.id)
        elif isinstance(value, ast.Attribute):
            processed_values.append(f"self.{extract_attribute(value)}")
    return processed_values

def extract_method_source_and_remove_tab(method: Callable)->str:
    """Extract the source code of a method and remove the tab

    :param method: The method
    :type method: Callable
    :return: The source code of the method
    :rtype: str
    """
    source = inspect.getsource(method)
    lines = source.split("\n")
    non_empty_lines = [line for line in lines if line.strip()]
    if non_empty_lines:
        first_line = non_empty_lines[0]
        indent = len(first_line) - len(first_line.lstrip())
        source_code = "\n".join([line[indent:] for line in lines])
    else:
        source_code = ""
    return source_code

def get_var_num(var):
    """Get the max number of a variable

    :param var: the variable
    :type var: str
    :return: the max number of the variable
    :rtype: int
    """
    #Check if var contains a number
    if not var.split("_")[-1].isdigit():
        return 0
    return int(var.split("_")[-1])

def get_max_var_num(vars):
    """Get the max number of a list of variables

    :param vars: the list of variables
    :type vars: list
    :return: the max number of the variables
    :rtype: int
    """
    if not len(vars):
        return 0
    return max([get_var_num(var) for var in vars])

def parse_assignment(node, assignments, cls):
    targets = node.targets
    for target in targets:
        process_targets(target, node, assignments, cls)

def process_targets(target, node, assignments, cls):
    # if isinstance(target, ast.Name):
    #         var_name = target.id
    #         process_var_name(var_name, node, assignments)
    # elif isinstance(target, ast.Attribute):
    #     var_name = f"self.{extract_attribute(target)}"
    #     process_var_name(var_name, node, assignments)
    # elif isinstance(target, ast.Tuple):
    #     for elt in target.elts:
    #         process_targets(elt, node, assignments)
    # else:
    #     raise Exception(f"Unsupported target type: {type(target)}")
    var_name = get_value(target, cls)
    process_var_name(var_name, node, assignments, cls)

def process_value(value:str, var_name:str, assignments:Dict[str,str]):
    vals = str(value).split(" ")
    for val in vals:
        max_num = get_max_var_num(assignments.keys())
        if max_num:
            val = f'{val}_{max_num}'
        if val in assignments.keys() and 'self.' not in val:
            try:
                value = value.replace(val, assignments[val])
            except TypeError as e:
                value = value.replace(val, str(assignments[val]))
            del assignments[val]
    # here is where I distinguish left from right
    assignments[var_name] = value

def process_var_name(var_name:str, node:ast.AST, assignments:Dict[str,str], cls:object):
    """Process the variable name

    :param var_name: the variable name
    :type var_name: str
    :param node: the node
    :type node: ast.AST
    """
    #Check if var_name in assignments
    #Need to handle repeated var_names that are attributes.
    if var_name in assignments.keys():
        #Add _i to var name
        var_name = f"{var_name}_{get_max_var_num(assignments.keys())+1}"
    value = get_value(node.value, cls)
    process_value(value, var_name, assignments)

def get_variable_assignments(cls: object, method_name: str)->Dict[str,str]:
    """Get the variable assignments of a method

    :param cls: the class with the method
    :type cls: object
    :param method_name: the name of the method
    :type method_name: str
    :return: the variable assignments in the method
    :rtype: dict
    """
    method = getattr(cls, method_name)
    source_code = extract_method_source_and_remove_tab(method)
    tree = ast.parse(source_code)
    assignments = {}
    #Iterate through all nodes in ast and get assignments
    for node in ast.walk(tree):
        #check if node is an assignment and get the left and righthand side values
        if isinstance(node, ast.Assign):
           parse_assignment(node, assignments, cls)
        #check if if statement
        elif isinstance(node, ast.If):
            #pull value out of if statement e.g if self.terminal: retrieve self.terminal
            value = get_value(node.test, cls)
            #generate 16 character random string
            random_string = ''.join(random.choices(string.ascii_uppercase + string.digits, k=16))
            process_value(value, random_string, assignments)
            #check conditional of if statement
            # if isinstance(node.test, ast.Compare):
            #     pass
            #check if node has an orelse
            # if node.orelse:
            #     #check if orelse is an assignment
            #     if isinstance(node.orelse[0], ast.Assign):
            #         parse_assignment(node.orelse[0], assignments, cls)
        elif isinstance(node, ast.While):
            #pull value out of if statement e.g if self.terminal: retrieve self.terminal
            value = get_value(node.test, cls)
            #generate 16 character random string
            random_string = ''.join(random.choices(string.ascii_uppercase + string.digits, k=16))
            process_value(value, random_string, assignments)
    return assignments  

def replace_intermediate_variables(attributes: Dict[str,str])->Dict[str,str]:
    """Replace intermediate variables in attributes

    :param attributes: the attributes
    :type attributes: Dict[str,str]
    :return: the attributes with intermediate variables replaced
    :rtype: Dict[str,str]
    """    
    intermediate_keys = []
    for key, value in attributes.items():
        if value in attributes.keys():
            attributes[key] = attributes[value]
            intermediate_keys.append(value)
    for key in intermediate_keys:
        del attributes[key]
    return attributes

def remove_self_and_underscore(word: str)->str:
    """Remove self and underscore after self dot

    :param word: the word
    :type word: str
    :return: the word without self and underscore after self dot
    :rtype: str
    """
    if "self." in word:
        #replace first underscore
        return word.replace("self.", "").split(".")[0]
        # if word[0] == "_":
        #     return word[1:]
        # else:
        #     return word
    else:
        return word

#extract all attributes from values in assignments
def extract_all_attributes_from_values_in_assignments(assignments: Dict[str,str])->Dict[str,Set[str]]:
    """Extract all attributes from values in assignments

    :param assignments: the assignments dictionary
    :type assignments: Dict[str,str]
    :return: assignments divided into inputs and outputs
    :rtype: Dict[str,Set[str]]
    """
    attributes = {'inputs':[], 'outputs':[]}
    for key, value in assignments.items():
        if isinstance(value, str):
            if "self." in value:
                d_val = []
                for word in value.split(" "):
                    if "self." in word:
                        d_val.append(remove_self_and_underscore(word))
                attributes['inputs'] += d_val
                attributes['outputs'].append(remove_self_and_underscore(key))
    attributes['inputs'] = list(set(attributes['inputs']))
    attributes['outputs'] = list(set(attributes['outputs']))
    return attributes

def process_words(word_str: str, split_val: str)->List[str]:
    """Process words

    :param word_str: the word string
    :type word_str: str
    :param split_val: the split value
    :type split_val: str
    :return: the processed words
    :rtype: List[str]
    """
    words = []
    for word in word_str.split(split_val):
        if "self." in word:
            words.append(remove_self_and_underscore(word))
    return words

def strip_all_but_class_attribute_names(assignments: Dict[str,str])->Dict[str,str]:
    """Strip all but class attribute names

    :param assignments: the assignments dictionary
    :type assignments: Dict[str,str]
    :return: assignments with only class attribute names
    :rtype: Dict[str,str]
    """
    stripped_assignments = {}
    for key, value in assignments.items():
        if 'self.' not in key:
            key = ''.join(random.choices(string.ascii_uppercase + string.digits, k=16))
        if ", " in key:
            key = tuple(process_words(key, ", "))
        else:
            key = (remove_self_and_underscore(key),)
        if isinstance(value, str):
            if ", " in value:
                words = tuple(process_words(value, ", "))
            else:
                words = process_words(value, " ")
            stripped_assignments[key] = words
        else:
            stripped_assignments[key] = ''
    return stripped_assignments

# def remove_non_io_attrs(stripped_assignments: Dict[str,str], cls: object)->Dict[str,str]:
#     """Remove non io attributes

#     :param assignments: the assignments dictionary
#     :type assignments: Dict[str,str]
#     :param io_attrs: the io attributes
#     :type io_attrs: List[str]
#     :return: the assignments with non io attributes removed
#     :rtype: Dict[str,str]
#     """
#     io_assignments = {}
#     for key, value in stripped_assignments.items():
#         if key in io_attrs:
#             stripped_assignments[key] = value
#     return stripped_assignments

def get_io_groupings(cls: object, method_name: str)->Dict[str,str]:
    """Get the input and output groupings of a method

    :param cls: class with method
    :type cls: object
    :param method_name: name of method
    :type method_name: str
    :return: input and output groupings of method
    :rtype: Dict[str,str]
    """    
    # assignments = replace_intermediate_variables(get_variable_assignments(cls, method_name))
    assignments = get_variable_assignments(cls, method_name)
    return strip_all_but_class_attribute_names(assignments)


def get_inputs_and_outputs(cls: object, method_name: str)->Dict[str,Set[str]]:
    """Get the inputs and outputs of a method

    :param cls: the class with the method
    :type cls: object
    :param method_name: the name of the method
    :type method_name: str
    :return: inputs and outputs of the method
    :rtype: Dict[str,Set[str]]
    """
    assignments = get_variable_assignments(cls, method_name)
    return extract_all_attributes_from_values_in_assignments(replace_intermediate_variables(assignments))

def dictify_list(container: Union[List[str], Dict[str, str], None])-> Dict[str, str]:
    """Convert a list to a dictionary with the same keys and values or return the dictionary

    :param container: The container to convert
    :type container: Union[List[str], Dict[str, str], None]
    :raises TypeError: if the container is not a list or dict
    :return: the converted container
    :rtype: Dict[str, str]
    """    
    if type(container) == list:
        return {val:val for val in container}
    elif type(container) == dict:
        return container
    elif container is None:
        return {}
    else:
        raise TypeError(f'{container} is not list or dict')


class IO:
    """A class to represent the inputs and outputs of a method
    """    
    def __init__(self, io:List[str], inst: 'AbstractComponent'):
        self._outputs = io[0]
        self._method = io[1]
        self._inputs = io[2]
        self._observer_names = []
        self._subject_names, out_rem = self.process_io(self._outputs, inst)
        self._outputs = [self.remove_underscore_number(x) for x in self._outputs if x not in out_rem]
        self._observer_names, inp_rem = self.process_io(self._inputs, inst)
        self._inputs = [self.remove_underscore_number(x) for x in self._inputs if x not in inp_rem]

    def process_io(self, inp_out:List[str], inst: 'AbstractComponent'):
        remove = []
        ret_list = []
        for val in inp_out:
            val = self.remove_underscore_number(val)
            if not hasattr(inst, val) and not hasattr(inst, f"_{val}") and val in inst._io_map:
                ret_list.append(inst._io_map[val])
            else: remove.append(val)
        return ret_list, remove
    
    def remove_underscore_number(self, val:str):
        # check if there is an underscore number at end of string using regex and remove it if there is
        return re.sub(r'_[0-9]+$', '', val)
    
    def __repr__(self) -> str:
        return f'IO({self._subject_names}, {self._observer_names})'

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, IO):
            return False
        return self._subject_names == __value._subject_names and self._observer_names == __value._observer_names and self._method == __value._method

class MasterIO:
    def __init__(self, inst):
        self.inst = inst
        self.io = []
        self.inputs = {}
        self.outputs = {}
        self.observers = {}
        self.subjects = {}
        self.getIO(inst)
        
    def getIO(self, inst):
        if not hasattr(inst, 'abstract_methods'):
            raise TypeError(f'{inst.__name__} does not have attribute abstract_methods')
        for method in inst.abstract_methods + inst._additional_io_methods:
            ios = get_io_groupings(inst, method)
            for key, value in ios.items():
                io = IO([key,method,value], inst)
                self.io.append(io)
                self.process_ios(io)
            self.remove_from_inputs_if_in_outputs()
        for io in self.io:
            self.create_obs_sub(io)

    def remove_from_inputs_if_in_outputs(self):
        for key in self.outputs.keys():
            if key in self.inputs.keys():
                del self.inputs[key]

    def process_ios(self, io:IO):
        for i, key in enumerate(io._outputs):
            if key not in self.outputs.keys():
                self.outputs[key] = io._subject_names[i]
            
        for i, value in enumerate(io._inputs):
            if value not in self.inputs.keys():
                self.inputs[value] = io._observer_names[i]

    def create_obs_sub(self, io:IO):
        for i, key in enumerate(io._outputs):
            if key not in self.subjects.keys():
                self.subjects[key] = Subject(io._subject_names[i], self.inst)
            
        for i, value in enumerate(io._inputs):
            if value not in self.outputs.keys():
                self.observers[value] = Observer(io._observer_names[i], self.inst)
    